# Copyright (C) 2015 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument

"""Tests for helpers.py."""

from unittest import mock

import pytest

import wikitcms.helpers as hl

# Only triplet_sort can handle these
HARD_SORTS = [
    ("9", "Beta", "RC1"),
    ("10", "Alpha", "TC1"),
    ("10", "Alpha", "TC2"),
    ("10", "Alpha", "TC10"),
    ("10", "Alpha", "RC1"),
]

# Weird old ones
OLD_SORTS = [
    ("12", "Alpha", "TCRegression"),
    ("12", "Beta", "PreBeta"),
    ("12", "Beta", "TC"),
    ("12", "Final", "Pre-RC"),
    ("12", "Final", "RC1"),
    ("21", "Rawhide", "2014 06"),
    ("21", "Nightly", "2014 08"),
    ("21", "Alpha", "TC1"),
]

# Standard stuff
STANDARD_SORTS = [
    ("22", "Final", "RC1"),
    ("22", "Postrelease", "20151015"),
    ("23", "Rawhide", "20150607"),
    ("23", "Branched", "20150717"),
    ("23", "Alpha", "TC1"),
    ("23", "Alpha", "TC2"),
    ("23", "Alpha", "RC1"),
    ("23", "Beta", "TC1"),
    ("23", "Beta", "TC2"),
    ("23", "Final", "RC1"),
    ("23", "Final", "RC1.1"),
    ("24", "Rawhide", "20160314.n.0"),
    ("24", "Rawhide", "20160314.n.1"),
    ("24", "Branched", "20160314.n.0"),
    ("24", "Branched", "20160318.n.0"),
    ("24", "Alpha", "1.1"),
    ("24", "Alpha", "1.2"),
    ("24", "Alpha", "2.1"),
    ("24", "Beta", "1.1"),
]


class TestHelpers:
    """Tests for the functions in helpers.py."""

    def test_fedora_release_sort(self):
        """Test for fedora_release_sort."""
        rels = OLD_SORTS + STANDARD_SORTS
        for (num, rel) in enumerate(rels[1:], 1):
            prevrel = rels[num - 1]
            assert hl.fedora_release_sort(" ".join(rel)) > hl.fedora_release_sort(" ".join(prevrel))

    def test_triplet_sort(self):
        """Test for triplet_sort and triplet_unsort."""
        rels = HARD_SORTS + OLD_SORTS + STANDARD_SORTS
        for (num, rel) in enumerate(rels[1:], 1):
            prevrel = rels[num - 1]
            sorttup = hl.triplet_sort(*rel)
            assert sorttup > hl.triplet_sort(*prevrel)
            assert hl.triplet_unsort(*sorttup) == (rel)

    def test_rreplace(self):
        """Tests for rreplace."""
        foos = "foofoofoo"
        bars = "barbarbar"
        assert hl.rreplace(bars, "bar", "foo", 1) == "barbarfoo"
        assert hl.rreplace(bars, "bar", "foo", 2) == "barfoofoo"
        assert hl.rreplace(bars, "bar", "foo", 3) == "foofoofoo"
        assert hl.rreplace(bars, "bar", "foo", 4) == "foofoofoo"
        assert hl.rreplace(foos, "bar", "foo", 1) == "foofoofoo"
        assert hl.rreplace(foos, "bar", "foo", 3) == "foofoofoo"

    def test_normalize(self):
        """Tests for normalize."""
        wikitext = "Foo_Bar Moo"
        oktext = "foo_bar_moo"
        assert hl.normalize(wikitext) == oktext
        assert hl.normalize(oktext) == oktext

    def test_find_bugs(self):
        """Test for find_bugs."""
        # text taken from a real Test Day page -
        # 2011-03-24_Power_Management - with a [[rhbug link
        # added as I couldn't find a page with all three. Some non-bug
        # text to make sure it's not erroneously processed, and an
        # example-type bug to make sure it's filtered
        text = """
| {{result|warn}} [https://bugzilla.redhat.com/show_bug.cgi?id=690177 switching of profile waiting forever] 
| [http://www.smolts.org/client/show/pub_1c75a13c-b42f-47b8-b886-a4ebf2c41305 HW]
| {{result|pass}} [https://fedoraproject.org/wiki/File:Pm-test-day-20110324-pub_54d92a30-b0cc-46ed-991c-00b3117ddb3a.txt bugreport]
| {{result|warn}} <ref> {{bz|690194}} - Probably what Albertpool is seeing</ref>
| {{result|warn}} <ref> [[rhbug:1239865|some rhbug link]]
{{bz|123456}} example bug
"""
        assert hl.find_bugs(text) == set(("690177", "690194", "1239865"))

    # this avoids needing network access and the value changing
    @mock.patch("fedfind.helpers.get_current_release", return_value=23, autospec=True)
    # this avoids needing network access
    @mock.patch("fedfind.release.Production.cid", "Fedora-24-20160314.1")
    @mock.patch("fedfind.release.Production.label", "Alpha-1.2")
    @mock.patch("fedfind.release.ModularProduction.cid", "Fedora-Modular-27-20171108.2")
    @mock.patch("fedfind.release.ModularProduction.label", "Beta-1.5")
    def test_cid_to_event(self, fakecurr):
        """Tests for cid_to_event."""
        bran = hl.cid_to_event("Fedora-24-20160314.n.2")
        assert bran == ("Fedora", "24", "Branched", "20160314.n.2")
        modbran = hl.cid_to_event("Fedora-Modular-27-20171110.n.1")
        assert modbran == ("Fedora-Modular", "27", "Branched", "20171110.n.1")
        rawh = hl.cid_to_event("Fedora-Rawhide-20160314.n.1")
        assert rawh == ("Fedora", "24", "Rawhide", "20160314.n.1")
        prod = hl.cid_to_event("Fedora-24-20160314.1")
        assert prod == ("Fedora", "24", "Alpha", "1.2")
        modprod = hl.cid_to_event("Fedora-Modular-27-20171108.2")
        assert modprod == ("Fedora-Modular", "27", "Beta", "1.5")
        # check that we fail intentionally on non-'Fedora' composes
        with pytest.raises(ValueError):
            hl.cid_to_event("Fedora-Cloud-28-20180711.0")
        # check that we fail intentionally on updates-ut composes
        with pytest.raises(ValueError):
            hl.cid_to_event("Fedora-27-updates-testing-20180711.0")
        with pytest.raises(ValueError):
            hl.cid_to_event("Fedora-27-updates-20180711.0")

    @mock.patch("fedfind.helpers.get_current_release", return_value=23, autospec=True)
    @mock.patch("fedfind.release.Production.cid", "Fedora-24-20160314.1")
    @mock.patch("fedfind.release.Production.label", "")
    def test_cid_to_event_fail(self, fakecurr):
        """Test to check cid_to_event fails if we cannot determine
        label for production compose.
        """
        with pytest.raises(ValueError):
            hl.cid_to_event("Fedora-24-20160314.1")


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2020 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument
# pylint: disable=invalid-name, too-few-public-methods

"""Tests for page.py."""

import datetime
import difflib
from decimal import Decimal
import json
import os
from unittest import mock

import mwclient.errors
import pytest

import wikitcms.event
import wikitcms.exceptions
import wikitcms.page
import wikitcms.result
import wikitcms.wiki

DATAPATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")


def _diff(old, new):
    """Convenience function for diffing in add_results tests."""
    diff = list(difflib.ndiff(old.splitlines(), new.splitlines()))
    # we enumerate so we can check the changes are in the
    # correct place; obviously this relies on the known
    # content of the sample data
    return [item for item in enumerate(diff) if item[1].startswith(("+", "-", "?"))]


@pytest.mark.usefixtures("fakemwp")
class TestPage:
    """Tests for Page instances."""

    # the "fjajah" is just to make sure we're running offline; if I
    # screw up and add a test that hits the network it"ll cause the
    # tests to hang/fail instead of succeeding a bit slower
    site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)

    def test_page_init(self):
        """Tests for Page.__init__."""
        pg = wikitcms.page.Page(self.site, "QA:Somepage")
        # __init__ stuff
        assert pg.checkname == "QA:Somepage"
        assert pg._sections is None
        assert pg.results_separators == []

    def test_page_sections(self, fakeapisections):
        """Tests for Page.sections."""
        pg = wikitcms.page.Page(self.site, "QA:Somepage")
        assert len(pg.sections) == 13
        assert pg.sections[0]["line"] == "Which tests to run"
        assert pg.sections[12]["line"] == "Upgrade tests"
        assert fakeapisections.call_count == 1
        # test caching
        assert len(pg.sections) == 13
        assert fakeapisections.call_count == 1
        # test error case
        pg._sections = None
        fakeapisections.side_effect = mwclient.errors.APIError(1, "foo", {})
        assert pg.sections == []
        assert fakeapisections.call_count == 2
        fakeapisections.side_effect = None
        # test cache clear, using simple return value for convenience
        fakeapisections.return_value = {"parse": {"sections": ["moo", "meep"]}}
        # until we save, changed API return value shouldn"t affect us
        assert pg.sections == []
        assert fakeapisections.call_count == 2
        with mock.patch("mwclient.page.Page.save", autospec=True):
            pg.save("foo", summary="bar")
        # now we saved, call count should increase and we should get
        # new values
        assert pg.sections == ["moo", "meep"]
        assert fakeapisections.call_count == 3

    @mock.patch("mwclient.page.Page.revisions")
    def test_page_creation_date(self, fakerev):
        """Tests for Page.creation_date."""
        pg = wikitcms.page.Page(self.site, "QA:Somepage")
        # the API gives us a timetuple, here
        date = datetime.datetime(2020, 3, 4).timetuple()
        fakerev.return_value = iter([{"timestamp": date}])
        assert pg.creation_date == "20200304"
        # this is a cached property, so we need to recreate the Page
        # instance now
        pg = wikitcms.page.Page(self.site, "QA:Somepage")
        fakerev.return_value = iter([])
        assert pg.creation_date == ""

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_page_write(self, fakesave):
        """Tests for Page.write."""
        pg = wikitcms.page.Page(self.site, "QA:Somepage")
        # need seedtext and summary or else it should raise
        with pytest.raises(ValueError):
            pg.write()
        pg.seedtext = "something"
        with pytest.raises(ValueError):
            pg.write()
        pg.seedtext = None
        pg.summary = "summary"
        with pytest.raises(ValueError):
            pg.write()
        pg.seedtext = "seedtext"
        pg.write()
        assert fakesave.call_count == 1
        # [0][0] is self
        assert fakesave.call_args[0][1] == "seedtext"
        assert fakesave.call_args[0][2] == "summary"
        assert fakesave.call_args[1]["createonly"] is True
        pg.write(createonly=False)
        assert fakesave.call_count == 2
        assert fakesave.call_args[1]["createonly"] is False
        pg.write(createonly=None)
        assert fakesave.call_count == 3
        assert fakesave.call_args[1]["createonly"] is None

    @mock.patch("time.sleep", autospec=True)
    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_page_save(self, fakesave, faketime):
        """Tests for Page.save."""
        pg = wikitcms.page.Page(self.site, "QA:Somepage")
        assert pg.save("foo", "summary", oldtext="foo") == {"nochange": ""}
        assert fakesave.call_count == 0
        pg.save("foo", "summary", oldtext="bar")
        assert fakesave.call_count == 1
        assert "oldtext" not in fakesave.call_args[1]
        # test the retry-once-on-fail behaviour by having parent save
        # raise EditError once, then go back to normal
        fakesave.side_effect = [mwclient.errors.EditError(), None]
        pg.save("foo", "summary")
        assert faketime.call_count == 1
        assert fakesave.call_count == 3


@pytest.mark.usefixtures("fakemwp")
class TestValidationPages:
    """Tests for ValidationPage instances (all the child classes)."""

    site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)

    def test_init(self):
        """Check all the various attributes of a ValidationPage set up
        during __init__ are correct for the common page types.
        """
        # a typical nightly page (should *not* be modular)
        nightly = wikitcms.page.NightlyPage(
            self.site, "32", "Installation", "Rawhide", "20200122.n.0"
        )
        # modular nightly page (based on the only nightly modular event
        # that ever existed...)
        modnightly = wikitcms.page.NightlyPage(
            self.site, "27", "Base", "Branched", "20171123.n.1", dist="Fedora-Modular"
        )
        # typical candidate page (should *not* be modular)
        candidate = wikitcms.page.ComposePage(self.site, "32", "Desktop", "RC", "1.2")
        # modular candidate page (based on the only modular compose we
        # ever did...)
        modcandidate = wikitcms.page.ComposePage(
            self.site, "27", "Server", "Beta", "1.5", dist="Fedora-Modular"
        )

        # IoT 'nightly'
        iotnightly = wikitcms.page.NightlyPage(
            self.site, "33", "General", "RC", "20200513.0", dist="Fedora-IoT"
        )
        assert nightly.release == "32"
        assert nightly.milestone == "Rawhide"
        assert nightly.compose == "20200122.n.0"
        assert nightly.version == "32 Rawhide 20200122.n.0"
        assert nightly.testtype == "Installation"
        assert nightly.dist == "Fedora"
        assert nightly.checkname == "Test Results:Fedora 32 Rawhide 20200122.n.0 Installation"
        assert nightly.summary == (
            "Relval bot-created Installation validation results page for "
            "Fedora 32 Rawhide 20200122.n.0"
        )
        assert nightly.results_separators == (
            "Test Matri",
            "Test Areas",
            "An unsupported test or configuration.  No testing is required.",
        )
        assert nightly.sortname == "32 100 20200122.n.0 Installation"
        assert nightly.sorttuple == (32, 100, Decimal("20200122.0"), "Installation")
        assert nightly.seedtext == (
            "{{subst:Validation results|testtype=Installation|release=32|milestone=Rawhide"
            "|date=20200122.n.0}}"
        )

        assert modnightly.release == "27"
        assert modnightly.milestone == "Branched"
        assert modnightly.compose == "20171123.n.1"
        assert modnightly.version == "27 Branched 20171123.n.1"
        assert modnightly.testtype == "Base"
        assert modnightly.dist == "Fedora-Modular"
        assert modnightly.checkname == "Test Results:Fedora-Modular 27 Branched 20171123.n.1 Base"
        assert modnightly.summary == (
            "Relval bot-created Base validation results page for "
            "Fedora-Modular 27 Branched 20171123.n.1"
        )
        assert modnightly.results_separators == (
            "Test Matri",
            "Test Areas",
            "An unsupported test or configuration.  No testing is required.",
        )
        assert modnightly.sortname == "27 150 20171123.n.1 Base"
        assert modnightly.sorttuple == (27, 150, Decimal("20171123.1"), "Base")
        assert modnightly.seedtext == (
            "{{subst:Modular validation results|testtype=Base|release=27|milestone=Branched"
            "|date=20171123.n.1}}"
        )

        assert candidate.release == "32"
        assert candidate.milestone == "RC"
        assert candidate.compose == "1.2"
        assert candidate.version == "32 RC 1.2"
        assert candidate.testtype == "Desktop"
        assert candidate.dist == "Fedora"
        assert candidate.checkname == "Test Results:Fedora 32 RC 1.2 Desktop"
        assert candidate.summary == (
            "Relval bot-created Desktop validation results page for " "Fedora 32 RC 1.2"
        )
        assert candidate.results_separators == (
            "Test Matri",
            "Test Areas",
            "An unsupported test or configuration.  No testing is required.",
        )
        # 6000 here is actually the value intended for "RC" as part of
        # a *compose* value, not "RC" as a *milestone* (with Pungi 4,
        # the GA milestone name changed from "Final" to "RC"). But it
        # doesn't seem to break anything, so I'm leaving it alone
        assert candidate.sortname == "32 6000 1.2 Desktop"
        assert candidate.sorttuple == (32, 800, Decimal("1.2"), "Desktop")
        assert candidate.seedtext == (
            "{{subst:Validation results|testtype=Desktop|release=32|milestone=RC|compose=1.2}}"
        )

        assert modcandidate.release == "27"
        assert modcandidate.milestone == "Beta"
        assert modcandidate.compose == "1.5"
        assert modcandidate.version == "27 Beta 1.5"
        assert modcandidate.testtype == "Server"
        assert modcandidate.dist == "Fedora-Modular"
        assert modcandidate.checkname == "Test Results:Fedora-Modular 27 Beta 1.5 Server"
        assert modcandidate.summary == (
            "Relval bot-created Server validation results page for " "Fedora-Modular 27 Beta 1.5"
        )
        assert modcandidate.results_separators == (
            "Test Matri",
            "Test Areas",
            "An unsupported test or configuration.  No testing is required.",
        )
        assert modcandidate.sortname == "27 400 1.5 Server"
        assert modcandidate.sorttuple == (27, 400, Decimal("1.5"), "Server")
        assert modcandidate.seedtext == (
            "{{subst:Modular validation results|testtype=Server|release=27|milestone=Beta"
            "|compose=1.5}}"
        )

        assert iotnightly.release == "33"
        assert iotnightly.milestone == "RC"
        assert iotnightly.compose == "20200513.0"
        assert iotnightly.version == "33 RC 20200513.0"
        assert iotnightly.testtype == "General"
        assert iotnightly.dist == "Fedora-IoT"
        assert iotnightly.checkname == "Test Results:Fedora-IoT 33 RC 20200513.0 General"
        assert iotnightly.summary == (
            "Relval bot-created General validation results page for Fedora-IoT 33 RC 20200513.0"
        )
        assert iotnightly.results_separators == (
            "Test Matri",
            "Test Areas",
            "An unsupported test or configuration.  No testing is required.",
        )
        assert iotnightly.sortname == "33 6000 20200513.0 General"
        assert iotnightly.sorttuple == (33, 800, Decimal("20200513.0"), "General")
        assert iotnightly.seedtext == (
            "{{subst:IoT validation results|testtype=General|release=33|milestone=RC"
            "|date=20200513.0}}"
        )

    def test_results_sections(self, fakeapisections):
        """Test the results_sections property."""
        # we aren"t using fakepages here so this doesn"t really matter
        # but let's stay consistent. we use fakeapisections not
        # fakepages as it's easier to adjust the return value
        page = wikitcms.page.NightlyPage(self.site, "32", "Server", "Branched", "20200322.n.0")
        assert [sec["line"] for sec in page.results_sections] == [
            "<b>Test Matrix</b>",
            "General tests",
            "FreeIPA server installation and functionality tests",
            "Domain joining tests: <b>FreeIPA</b>",
            "Domain joining tests: <b>Active Directory</b>",
            "Domain client tests",
            "PostgreSQL server installation and functionality tests",
            "Upgrade tests",
        ]
        # the code has a special case for handling a case where we
        # don"t positively find a first result section but do find a
        # section named "Key", but I don"t remember what that's for
        # and can"t find a real example (the oldest page we handle,
        # Test_Results:Fedora_12_Alpha_RC1_Install , has a
        # "Test Matrix" section), so let's make one up by dropping
        # the "Test Matrix" section from our test data
        del fakeapisections.return_value["parse"]["sections"][5]
        assert [sec["line"] for sec in page.results_sections] == [
            "General tests",
            "FreeIPA server installation and functionality tests",
            "Domain joining tests: <b>FreeIPA</b>",
            "Domain joining tests: <b>Active Directory</b>",
            "Domain client tests",
            "PostgreSQL server installation and functionality tests",
            "Upgrade tests",
        ]

    def test_get_resultrows(self, fakepages):
        """Test the get_resultrows() method."""
        # use one of the pages supported by "fakepages"
        page = wikitcms.page.NightlyPage(self.site, "32", "Server", "Branched", "20200322.n.0")
        rrows = page.get_resultrows()
        assert len(rrows) == 26
        # just sample a couple of rows
        assert rrows[0].testcase == "QA:Testcase_kickstart_firewall_disabled"
        assert rrows[13].name == "QA:Testcase_realmd_join_kickstart"
        assert rrows[25].secid == "8"

    def test_find_resultrow(self, fakepages):
        """Test find_resultrow() method, including various awkward
        cases.
        """
        page = wikitcms.page.NightlyPage(self.site, "32", "Server", "Branched", "20200322.n.0")
        # simple case: one row found by test case
        row = page.find_resultrow(testcase="QA:Testcase_kickstart_firewall_disabled")
        assert row.testcase == "QA:Testcase_kickstart_firewall_disabled"
        assert row.secid == "2"
        # should raise TooManyError for test case found more than once
        # with different names
        with pytest.raises(wikitcms.exceptions.TooManyError):
            row = page.find_resultrow(testcase="QA:Testcase_domain_client_authenticate")
        # ...but if we give a name, should be fine
        row = page.find_resultrow(
            testcase="QA:Testcase_domain_client_authenticate", testname="FreeIPA"
        )
        # and in fact just searching on the name should also work, if
        # we give the *exact* name
        rowagain = page.find_resultrow(testname="(FreeIPA)")
        assert rowagain.matches(row)
        # should raise TooManyError for test case found more than once
        # with different sections
        with pytest.raises(wikitcms.exceptions.TooManyError):
            row = page.find_resultrow(testcase="QA:Testcase_realmd_join_kickstart")
        # ...but if we give section, should be fine
        row = page.find_resultrow(testcase="QA:Testcase_realmd_join_kickstart", section="FreeIPA")
        # should raise NotFoundError if we find nothing
        with pytest.raises(wikitcms.exceptions.NotFoundError):
            row = page.find_resultrow(testcase="QA:Testcase_non_existent")
        # test case where 2+ test cases have the search string in them
        # but one matches it exactly; we should find that row and not
        # raise TooManyError or anything
        row = page.find_resultrow(testcase="QA:Testcase_freeipa_replication")
        assert row.testcase == "QA:Testcase_freeipa_replication"

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_update_current(self, fakesave):
        """Test for update_current() method."""
        page = wikitcms.page.NightlyPage(self.site, "32", "Server", "Branched", "20200322.n.0")
        page.update_current()
        assert fakesave.call_count == 1
        args = fakesave.call_args[0]
        # this is the Page instance itself
        assert args[0].name == "Test Results:Current Server Test"
        assert args[1] == "#REDIRECT [[Test Results:Fedora 32 Branched 20200322.n.0 Server]]"
        assert args[2] == "relval: update to current event"

        # now check modular case
        page = wikitcms.page.ComposePage(
            self.site, "32", "Server", "Beta", "1.1", dist="Fedora-Modular"
        )
        page.update_current()
        assert fakesave.call_count == 2
        args = fakesave.call_args[0]
        # this is the Page instance itself
        assert args[0].name == "Test Results:Current Modular Server Test"
        assert args[1] == "#REDIRECT [[Test Results:Fedora-Modular 32 Beta 1.1 Server]]"
        assert args[2] == "relval: update to current event"

        # IoT case
        page = wikitcms.page.NightlyPage(
            self.site, "33", "General", "RC", "20200513.0", dist="Fedora-IoT"
        )
        page.update_current()
        assert fakesave.call_count == 3
        args = fakesave.call_args[0]
        # this is the Page instance itself
        assert args[0].name == "Test Results:Current IoT General Test"
        assert args[1] == "#REDIRECT [[Test Results:Fedora-IoT 33 RC 20200513.0 General]]"
        assert args[2] == "relval: update to current event"


@pytest.mark.usefixtures("fakemwp", "fakepages")
class TestAddResults:
    """Tests for ValidationPage.add_results() method. Is its own class
    so these tests can share some stuff.
    """

    @pytest.fixture(autouse=True)
    def common_setup(self, request, fakemwp, fakepages):
        """This seems like a kinda baroque way to do things, but the
        problem is we need to use the fakemwp and fakepages fixtures
        *in this code*, and I can"t figure a way to do that while just
        setting these things up as class attributes as we ordinarily
        would. This autouse fixture which sets attributes on the
        request class seems to work instead...
        """
        us = request.cls
        us.site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        us.page = wikitcms.page.NightlyPage(us.site, "32", "Server", "Branched", "20200322.n.0")
        us.text = us.page.text()
        us.row = us.page.find_resultrow(testcase="QA:Testcase_kickstart_firewall_disabled")
        us.row2 = us.page.find_resultrow(testcase="QA:Testcase_FreeIPA_realmd_login")
        us.passed = wikitcms.result.Result("pass", "wikitcms")
        us.failed = wikitcms.result.Result("fail", "wikitcms")

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_simple(self, fakesave, fakepages):
        """Simple one result test."""
        # pylint: disable=no-member
        ret = self.page.add_results({self.row: [("x86_64", self.passed)]})
        # should be no dupes
        assert ret == []
        assert fakesave.call_count == 1
        # summary
        expsumm = "Result(s) for test(s): QA:Testcase_kickstart_firewall_disabled filed via relval"
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (14, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_multiple(self, fakesave, fakepages):
        """Multiple result test. Also tests overwriting
        {{result|none}}.
        """
        # pylint: disable=no-member
        ret = self.page.add_results({self.row: [("x86_64", self.passed), ("aarch64", self.failed)]})
        assert ret == []
        assert fakesave.call_count == 1
        expsumm = "Result(s) for test(s): QA:Testcase_kickstart_firewall_disabled filed via relval"
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (14, "- | {{result|none}}"),
            (15, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
            (16, "+ | {{result|fail|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_multiple_row(self, fakesave, fakepages):
        """Multiple row, different sections test."""
        # pylint: disable=no-member
        ret = self.page.add_results(
            {self.row: [("x86_64", self.passed)], self.row2: [("Result", self.failed)]}
        )
        assert ret == []
        assert fakesave.call_count == 1
        # the ordering of test cases in the summary is reversed as we
        # reverse the dict to ensure editing accuracy...
        expsumm = (
            "Result(s) for test(s): QA:Testcase_FreeIPA_realmd_login, "
            "QA:Testcase_kickstart_firewall_disabled filed via relval"
        )
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (14, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
            (139, "- | {{result|pass|coconut|bot=true}}"),
            (140, "+ | {{result|pass|coconut|bot=true}}{{result|fail|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_single_env(self, fakesave, fakepages):
        """If there's only one env, we use it even if the passed env
        doesn"t match at all (this is an arguable choice but it's what
        we"ve done for years...)
        """
        # pylint: disable=no-member
        self.page.add_results({self.row2: [("bananas", self.passed)]})
        assert fakesave.call_count == 1
        expsumm = "Result(s) for test(s): QA:Testcase_FreeIPA_realmd_login filed via relval"
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (138, "- | {{result|pass|coconut|bot=true}}"),
            (139, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_incomplete_env(self, fakesave, fakepages):
        """An incomplete env string match is expected to work if it is
        unique.
        """
        # pylint: disable=no-member
        ret = self.page.add_results({self.row: [("x86", self.passed)]})
        assert ret == []
        assert fakesave.call_count == 1
        # summary
        expsumm = "Result(s) for test(s): QA:Testcase_kickstart_firewall_disabled filed via relval"
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (14, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_incomplete_dupes(self, fakesave, fakepages):
        """Duplicate submissions (when we have a result in same cell
        from same reporter) should be returned, as a list of tuples
        (row, env, Result) where the Result is the *submitted* result
        not the existing one.
        """
        # pylint: disable=no-member
        cocores = wikitcms.result.Result("fail", "coconut", bot=True)
        ret = self.page.add_results(
            {
                self.row: [("x86_64", cocores), ("x86_64", self.passed)],
                self.row2: [("Result", cocores)],
            }
        )
        # reversed, as results dict is reversed by add_results...
        assert ret == [(self.row2, "Result", cocores), (self.row, "x86_64", cocores)]
        # the row we didn't touch as we only had dupe results should be
        # left out of the summary text
        expsumm = "Result(s) for test(s): QA:Testcase_kickstart_firewall_disabled filed via relval"
        assert fakesave.call_args[0][2] == expsumm
        # the dupe results should *not* be posted
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (14, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

        # now check with allowdupe=True
        ret = self.page.add_results(
            {
                self.row: [("x86_64", cocores), ("x86_64", self.passed)],
                self.row2: [("Result", cocores)],
            },
            allowdupe=True,
        )
        assert ret == []
        expsumm = (
            "Result(s) for test(s): QA:Testcase_FreeIPA_realmd_login, "
            "QA:Testcase_kickstart_firewall_disabled filed via relval"
        )
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (
                14,
                "+ | {{result|pass|coconut|bot=true}}"
                "{{result|fail|coconut|bot=true}}{{result|pass|wikitcms}}",
            ),
            (139, "- | {{result|pass|coconut|bot=true}}"),
            (140, "+ | {{result|pass|coconut|bot=true}}{{result|fail|coconut|bot=true}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_results_ellipsization(self, fakesave, fakepages):
        """Test the summary ellipsization when more than three test
        cases have results.
        """
        # pylint: disable=no-member
        row3 = self.page.find_resultrow(testcase="QA:Testcase_postgresql_server_installation")
        row4 = self.page.find_resultrow(
            testcase="QA:Testcase_upgrade_dnf_current_server_domain_controller"
        )
        ret = self.page.add_results(
            {
                self.row: [("x86_64", self.passed)],
                self.row2: [("Result", self.passed)],
                row3: [("x86_64", self.passed)],
                row4: [("x86_64", self.passed)],
            }
        )
        assert ret == []
        # Reverse order again...
        expsumm = (
            "Result(s) for test(s): QA:Testcase_upgrade_dnf_current_server_domain_"
            "controller, QA:Testcase_postgresql_server_installation, "
            "QA:Testcase_FreeIPA_realmd_login... filed via relval"
        )
        assert fakesave.call_args[0][2] == expsumm

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_add_result(self, fakesave, fakepages):
        """As add_result is just a thin wrapper for add_results (to
        retain backward compatibility from before we added multiple
        simultaneous result submission), let's test it here.
        """
        # pylint: disable=no-member
        # this does the same as test_add_results_simple
        ret = self.page.add_result(self.passed, self.row, "x86_64")
        # should be no dupes
        assert ret == []
        assert fakesave.call_count == 1
        # summary
        expsumm = "Result(s) for test(s): QA:Testcase_kickstart_firewall_disabled filed via relval"
        assert fakesave.call_args[0][2] == expsumm
        expdiff = [
            (13, "- | {{result|pass|coconut|bot=true}}"),
            (14, "+ | {{result|pass|coconut|bot=true}}{{result|pass|wikitcms}}"),
        ]
        assert _diff(self.text, fakesave.call_args[0][1]) == expdiff


@pytest.mark.usefixtures("fakemwp")
class TestSummaryPage:
    """Test for the SummaryPage class."""

    @mock.patch("fedfind.release.Compose.exists", True)
    @mock.patch("fedfind.release.Compose.all_images", ["foo"])
    @mock.patch("wikitcms.wiki.Wiki.get_testtypes", return_value=["Installation", "Base"])
    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_summary_page(self, fakesave, fakeget):
        """General tests for SummaryPage. It's a simple class."""
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        event = wikitcms.event.ComposeEvent(site, "32", "Beta", "1.2", cid="Fedora-32-20200312.0")
        summ = event.summary_page
        assert isinstance(summ, wikitcms.page.SummaryPage)
        assert summ.checkname == "Test Results:Fedora 32 Beta 1.2 Summary"
        # checking the summary summary, like whoah, dude
        assert summ.summary == (
            "Relval bot-created validation results summary for " "Fedora 32 Beta 1.2"
        )
        assert summ.seedtext == (
            "Fedora 32 Beta 1.2 [[QA:Release validation test plan|release "
            "validation]] summary. This page shows the results from all the "
            "individual result pages for this compose together. You can file "
            "results directly from this page and they will be saved into the "
            "correct individual result page. To see test instructions, visit "
            "any of the individual pages (the section titles are links). You "
            "can find download links below.\n\n__TOC__\n\n== Downloads ==\n{{"
            "Fedora 32 Beta 1.2 Download}}\n\n== [[Test Results:Fedora 32 Beta 1.2 Installation|"
            "Installation]] ==\n{{Test Results:Fedora 32 Beta 1.2 Installation}}\n\n== "
            "[[Test Results:Fedora 32 Beta 1.2 Base|Base]] ==\n{{Test Results:Fedora 32 Beta 1.2 "
            "Base}}"
        )
        summ.update_current()
        assert fakesave.call_count == 1
        args = fakesave.call_args[0]
        # this is the Page instance itself
        assert args[0].name == "Test Results:Current Summary"
        assert args[1] == "#REDIRECT [[Test Results:Fedora 32 Beta 1.2 Summary]]"
        assert args[2] == "relval: update to current event"
        fakesave.reset_mock()

        # modular checks
        modevent = wikitcms.event.ComposeEvent(
            site, "27", "Beta", "1.5", cid="Fedora-Modular-27-20171108.2", dist="Fedora-Modular"
        )
        summ = modevent.summary_page
        assert isinstance(summ, wikitcms.page.SummaryPage)
        assert summ.checkname == "Test Results:Fedora-Modular 27 Beta 1.5 Summary"
        assert summ.summary == (
            "Relval bot-created validation results summary for Fedora-Modular 27 Beta 1.5"
        )
        # let's not test the whole goddamned thing again
        assert summ.seedtext.startswith("Fedora-Modular 27 Beta 1.5 [[")
        summ.update_current()
        assert fakesave.call_count == 1
        args = fakesave.call_args[0]
        # this is the Page instance itself
        assert args[0].name == "Test Results:Current Modular Summary"
        assert args[1] == "#REDIRECT [[Test Results:Fedora-Modular 27 Beta 1.5 Summary]]"
        assert args[2] == "relval: update to current event"
        fakesave.reset_mock()

        # IoT checks
        iotevent = wikitcms.event.NightlyEvent(site, "33", "RC", "20200513.0", dist="Fedora-IoT")
        summ = iotevent.summary_page
        assert isinstance(summ, wikitcms.page.SummaryPage)
        assert summ.checkname == "Test Results:Fedora-IoT 33 RC 20200513.0 Summary"
        assert summ.summary == (
            "Relval bot-created validation results summary for Fedora-IoT 33 RC 20200513.0"
        )
        # let's not test the whole goddamned thing again
        assert summ.seedtext.startswith("Fedora-IoT 33 RC 20200513.0 [[")
        summ.update_current()
        assert fakesave.call_count == 1
        args = fakesave.call_args[0]
        # this is the Page instance itself
        assert args[0].name == "Test Results:Current IoT Summary"
        assert args[1] == "#REDIRECT [[Test Results:Fedora-IoT 33 RC 20200513.0 Summary]]"
        assert args[2] == "relval: update to current event"
        fakesave.reset_mock()


@pytest.mark.usefixtures("fakemwp", "fakepages", "fakeimages")
class TestDownloadPage:
    """Tests for the DownloadPage class."""

    # we have to patch these to avoid network round trips in fedfind
    @mock.patch("fedfind.release.Compose.cid", "Fedora-32-20200312.0")
    @mock.patch("fedfind.release.Compose.exists", True)
    def test_download_page(self):
        """Test for DownloadPage. It's pretty hard to break this down
        into unit tests, given how it works, so we just test it all at
        once: we run it on the real data from a real compose, and
        check the output matches what we expect.
        """
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        event = wikitcms.event.ComposeEvent(site, "32", "Beta", "1.2", cid="Fedora-32-20200312.0")
        page = event.download_page
        assert page.checkname == "Template:Fedora 32 Beta 1.2 Download"
        assert page.summary == "Relval bot-created download page for Fedora 32 Beta 1.2"
        assert page.event is event
        # seedtext is very long so we stash the expected value in a
        # file and read it here
        exptextpath = os.path.join(DATAPATH, "Template:Fedora 32 Beta 1.2 Download.seedtext")
        with open(exptextpath, "r") as exptextfh:
            exptext = exptextfh.read()
        assert page.seedtext == exptext

        # modular checks
        modevent = wikitcms.event.ComposeEvent(
            site, "27", "Beta", "1.5", cid="Fedora-Modular-27-20171108.2", dist="Fedora-Modular"
        )
        page = modevent.download_page
        assert page.checkname == "Template:Fedora-Modular 27 Beta 1.5 Download"
        assert page.summary == "Relval bot-created download page for Fedora-Modular 27 Beta 1.5"
        # that's all we need to check for modular

        # IoT checks
        iotevent = wikitcms.event.NightlyEvent(site, "33", "RC", "20200513.0", dist="Fedora-IoT")
        page = iotevent.download_page
        assert page.checkname == "Template:Fedora-IoT 33 RC 20200513.0 Download"
        assert page.summary == "Relval bot-created download page for Fedora-IoT 33 RC 20200513.0"
        # that's all we need to check for IoT


@pytest.mark.usefixtures("fakemwp", "fakepages", "fakeimages")
class TestAMIPage:
    """Tests for the AMIPage class."""

    # we have to patch these to avoid network round trips in fedfind
    @mock.patch("fedfind.release.Compose.cid", "Fedora-32-20200312.0")
    @mock.patch("fedfind.release.Compose.exists", True)
    def test_ami_page(self):
        """Test for AMIPage. As with DownloadPage, it's hard to split
        this into unit tests, so we feed in real world data (more or
        less) and check the output is as expected.
        """
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        event = wikitcms.event.ComposeEvent(site, "32", "Beta", "1.2", cid="Fedora-32-20200312.0")
        page = event.ami_page
        assert page.checkname == "Template:Fedora 32 Beta 1.2 AMI"
        assert page.summary == "Relval bot-created AMI page for Fedora 32 Beta 1.2"
        assert page.event is event

        def fakedownloadjson(url):
            """Mock for fedfind.helpers.download_json to fake up a
            datagrepper query response used to find the AMIs. The data
            that backs this is close to but not quite 'real' data -
            it's the real message dicts only of the messages that were
            actually for this compose, plus *one* message that wasn't
            (to test that filtering those out works), with the headers
            and most of the additional data in the query response
            dropped. The *real* real data for the relevant two days
            is like 25 pages long, so we had to filter it. We hack up
            splitting it across two 'pages' to test the pagination
            handling.
            """
            page = 1
            if "page=2" in url:
                # dumb but enough
                page = 2
            path = os.path.join(DATAPATH, "Fedora-32-20200312.0.ami.dgdata.{0}.json".format(page))
            with open(path, "r") as amifh:
                amidata = json.load(amifh)
            return amidata

        # again, read expected seedtext from a file
        exptextpath = os.path.join(DATAPATH, "Template:Fedora 32 Beta 1.2 AMI.seedtext")
        with open(exptextpath, "r") as exptextfh:
            exptext = exptextfh.read()
        with mock.patch("fedfind.helpers.download_json", fakedownloadjson):
            assert page.seedtext == exptext


@pytest.mark.usefixtures("fakemwp")
class TestTestDayPage:
    """Tests for the TestDayPage class."""

    def test_init(self):
        """Tests for basic __init__ stuff."""
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        page = wikitcms.page.TestDayPage(site, "2020-03-20", "Cloud Testday")
        assert page.checkname == "Test Day:2020-03-20 Cloud Testday"
        assert page.date == "2020-03-20"
        assert page.subject == "Cloud Testday"
        assert page.results_separators == ("Test Results", "Results")

    def test_bugs(self, fakepages):
        """Tests for the bugs() property."""
        # this page uses the bugs-in-results-template style
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        page = wikitcms.page.TestDayPage(site, "2013-11-05", "Printing")
        assert page.bugs == ["1026909", "1026914", "1026928", "1026940", "1026949", "1027425"]
        # this one uses the app-generated bugs-as-references style
        # and has some direct Bugzilla URLs also, also has lots of
        # duplication between entries
        page = wikitcms.page.TestDayPage(site, "2016-10-24", "Cloud")
        assert page.bugs == ["1384150", "1387934", "1388000"]

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_fix_app_results(self, fakesave, fakepages):
        """Test for fix_app_results."""
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        page = wikitcms.page.TestDayPage(site, "2016-10-24", "Cloud")
        page.fix_app_results()
        assert fakesave.call_count == 1
        args = fakesave.call_args[0]
        kwargs = fakesave.call_args[1]
        assert kwargs["summary"] == (
            "Fix testday app-generated results to use " "{{result}} template for bug references"
        )
        # FIXME: we actually miss several conversions here as we do not
        # handle {{bz}} and comment text mixed up in a single ref tag
        # FIXME 2: I even had to cheat and edit the text of the page
        # slightly as there's a stray space between the end of the
        # template and the closing </ref>
        assert args[1] == page.text().replace(
            "{{result|fail}}<ref>{{bz|1388000}}</ref>", "{{result|fail||1388000}}"
        )

    @mock.patch("mwclient.page.Page.save", autospec=True)
    def test_long_refs(self, fakesave, fakepages):
        """Test for long_refs."""
        site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)
        page = wikitcms.page.TestDayPage(site, "2016-10-24", "Cloud")
        page.long_refs()
        # read expected  modified text from a file, it's too long to
        # inline
        exptextpath = os.path.join(DATAPATH, "Test_Day:2016-10-24_Cloud.longrefs.txt")
        with open(exptextpath, "r") as exptextfh:
            exptext = exptextfh.read()
        assert fakesave.call_args[0][1] == exptext
        # summary
        assert fakesave.call_args[1]["summary"] == (
            "Move long comments to a separate " "section at end of page"
        )



# Copyright (C) 2020 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument
# pylint: disable=invalid-name

"""Test fixtures, etc."""

from __future__ import unicode_literals
from __future__ import print_function

import json
import os
from unittest import mock

import pytest

DATAPATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")


@pytest.fixture
def fakemwp(request):
    """Fixture that mocks out mwclient.page.Page.__init__ as long
    as it's active.
    """

    def fakemwpinit(self, site, name, *args, **kwargs):
        """Stub init for mwclient.Page, we can't just mock it out as we
        need to set site and name (and for category generator testing,
        _info, though it can be None).
        """
        self.site = site
        # this is for testing TcmsGeneratorList's capability to handle
        # getting a page by its id
        if name == 80736:
            name = "Test Day:2019-12-09 Kernel 5.4 Test Week"
        self.name = name
        self._info = None

    with mock.patch("mwclient.page.Page.__init__", fakemwpinit):
        yield


@pytest.fixture
def fakeapisections(request):
    """Fixture that mocks out the API to give a response that's
    appropriate for a 'sections' call on a test result page, and
    yields the mock for modifications. Based on the real response for
    Test Results:Fedora 32 Branched 20200322.n.0 Server on 2020-03-25.
    """
    sectpath = os.path.join(
        DATAPATH, "Test_Results:Fedora_32_Branched_20200322.n.0_Server.sections.json"
    )
    with open(sectpath, "r") as sectfh:
        sectjson = json.load(sectfh)
    with mock.patch("wikitcms.wiki.Wiki.api", autospec=True) as fakeapi:
        fakeapi.return_value = {"parse": {"sections": sectjson}}
        yield fakeapi


@pytest.fixture
def fakepages(request):
    """Fixture that mocks out various Page attributes, methods etc.
    with return values read from data files based on the page name.
    The backing files contain real data from the real wiki.
    """

    def faketext(self, section=None, expandtemplates=False, cache=True, slot="main"):
        textpath = os.path.join(DATAPATH, self.name.replace(" ", "_")) + ".txt"
        with open(textpath, "r") as textfh:
            text = textfh.read()
        return text

    @property
    def fakesections(self):
        sectpath = os.path.join(DATAPATH, self.name.replace(" ", "_")) + ".sections.json"
        with open(sectpath, "r") as sectfh:
            sectjson = json.load(sectfh)
        return sectjson

    with mock.patch("mwclient.page.Page.text", faketext):
        with mock.patch("wikitcms.page.Page.sections", fakesections):
            yield


@pytest.fixture
def fakeimages(request):
    """Fixture that mocks out fedfind all_images with an appropriate
    value (from backing data) for the compose. Used when we need an
    event to have authentic image data.
    """

    @property
    def fakeallimages(self):
        imagespath = os.path.join(DATAPATH, self.cid) + ".images.json"
        with open(imagespath, "r") as imagesfh:
            imagesjson = json.load(imagesfh)
        return imagesjson

    with mock.patch("fedfind.release.Release.all_images", fakeallimages):
        yield



# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# I disagree, pylint! Shut up!
# pylint: disable=too-many-arguments, too-many-instance-attributes

"""Classes that describe test events."""

import abc
import logging

from cached_property import cached_property
import fedfind.helpers
import fedfind.release
import mwclient.errors

from wikitcms.exceptions import FedfindNotFoundError
import wikitcms.helpers
import wikitcms.listing
import wikitcms.page

logger = logging.getLogger(__name__)


class ValidationEvent:
    """A parent class for different types of release validation event.
    site must be an instance of wikitcms.Wiki, already with
    appropriate access rights for any actions to be performed (i.e.
    things instantiating an Event are expected to do site.login
    themselves if needed). Required attributes: shortver,
    category_page. "dist" lets us create multiple event streams for
    different compose streams: it defaults to "Fedora", which gives
    us event and page names for the main stream of composes and Fedora
    releases, but can be set to e.g. "Fedora IoT" for a validation
    event for an IoT compose. The provided string is simply included
    in page names, category names and so on.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, site, release, milestone="", compose="", dist="Fedora"):
        self.site = site
        self.release = release
        self.milestone = str(milestone)
        try:
            self.compose = fedfind.helpers.date_check(compose, fail_raise=True, out="str")
        except ValueError:
            self.compose = str(compose)
        self.dist = wikitcms.helpers.validate_dist(dist)
        self.version = f"{self.release} {self.milestone} {self.compose}"
        # Sorting helpers. sortname is a string, sorttuple is a
        # 4-tuple. sorttuple is more reliable. See the function docs.
        self.sortname = wikitcms.helpers.fedora_release_sort(self.version)
        self.sorttuple = wikitcms.helpers.triplet_sort(self.release, self.milestone, self.compose)

    @abc.abstractproperty
    def _current_content(self):
        """The content for the CurrentFedoraCompose template for
        this test event.
        """

    @abc.abstractproperty
    def _pagetype(self):
        """The ValidationPage class to be used for this event's pages
        (for use by valid_pages).
        """

    @abc.abstractproperty
    def category_page(self):
        """The category page for this event. Is a property because
        page instantiation requires a remote trip.
        """

    @property
    def result_pages(self):
        """A list of wikitcms page objects for currently-existing
        pages that are a part of this test event, according to the
        naming convention.
        """
        pages = self.site.allresults(prefix=f"{self.dist} {self.version} ")
        return [p for p in pages if isinstance(p, wikitcms.page.ValidationPage)]

    @property
    def download_page(self):
        """The DownloadPage for this event. Is a property because page
        instantiation requires a remote trip.
        """
        return wikitcms.page.DownloadPage(self.site, self, dist=self.dist)

    @property
    def ami_page(self):
        """The AMIPage for this event. Is a property because page
        instantiation requires a remote trip.
        """
        return wikitcms.page.AMIPage(self.site, self, dist=self.dist)

    @property
    def parent_category_page(self):
        """The parent category page for this event. Is a property for
        the same reason as download_page.
        """
        return wikitcms.listing.ValidationCategory(self.site, self.release, dist=self.dist)

    @property
    def valid_pages(self):
        """A list of the expected possible result pages (as
        page.ValidationPage objects) for this test event, derived from
        the available test types and the naming convention.
        """
        types = self.site.get_testtypes(dist=self.dist)
        return [
            self._pagetype(
                self.site,
                self.release,
                typ,
                milestone=self.milestone,
                compose=self.compose,
                dist=self.dist,
            )
            for typ in types
        ]

    @property
    def summary_page(self):
        """The page.SummaryPage object for the event's result summary
        page. Very simple property, but not set in __init__ as the
        summary page object does (slow) wiki roundtrips in  __init__.
        """
        return wikitcms.page.SummaryPage(self.site, self, dist=self.dist)

    @cached_property
    def ff_release(self):
        """A fedfind release object matching this event."""
        # note: fedfind has a hack that parses date and respin out
        # of a dot-separated compose, since wikitcms smooshes them
        # into the compose value.
        try:
            return fedfind.release.get_release(
                release=self.release, milestone=self.milestone, compose=self.compose, dist=self.dist
            )
        except ValueError as err:
            try:
                if self._cid:
                    return fedfind.release.get_release(cid=self._cid)
            except AttributeError:
                raise FedfindNotFoundError(err)
            raise FedfindNotFoundError(err)

    @property
    def ff_release_images(self):
        """A fedfind release object matching this event, that has
        images. If we can't find one, raise an exception. For the
        base class this just acts as a check on ff_release; it does
        something more clever in ComposeEvent.
        """
        rel = self.ff_release
        if rel.all_images:
            return rel
        err = f"Could not find fedfind release with images for event {self.version}"
        raise FedfindNotFoundError(err)

    def update_current(self):
        """Make the CurrentFedoraCompose template on the wiki point to
        this event. The template is used for the Current (testtype)
        Test redirect pages which let testers find the current results
        pages, and for other features of wikitcms/relval. Children
        must define _current_content.
        """
        content = "{{tempdoc}}\n<onlyinclude>{{#switch: {{{1|full}}}\n"
        content += self._current_content
        content += "}}</onlyinclude>\n[[Category: Fedora Templates]]"
        curr = self.site.pages[f"Template:Current{self.dist}Compose"]
        curr.save(content, "relval: update to current event", createonly=None)

    def create(self, testtypes=None, force=False, current=True, check=False):
        """Create the event, by creating its validation pages,
        summary page, download page, category pages, and updating the
        current redirects. 'testtypes' can be an iterable that limits
        creation to the specified testtypes. If 'force' is True, pages
        that already exist will be recreated (destroying any results
        on them). If 'current' is False, the current redirect pages
        will not be updated. If 'check' is true, we check if any
        result page already exists first, and bail immediately if so
        (so we don't start creating pages then hit one that exists and
        fail half-way through, for things that don't want that.) 'cid'
        can be set to a compose ID, this is for forcing the compose
        location at event creation time when we know we're not going
        to be able to find it any way and is a short-term hack that
        will be removed.
        """
        logger.info("Creating validation event %s", self.version)
        createonly = True
        if force:
            createonly = None
        pages = self.valid_pages
        if testtypes:
            logger.debug("Restricting to testtypes %s", " ".join(testtypes))
            pages = [pag for pag in pages if pag.testtype in testtypes]
        if not pages:
            raise ValueError("No result pages to create! Wrong test type?")
        if check:
            if any(pag.text() for pag in pages):
                raise ValueError("A result page already exists!")

        # NOTE: download page creation for ComposeEvents will only
        # work if:
        # * the compose has being synced to stage, OR
        # * the compose has been imported to PDC, OR
        # * you used get_validation_event and passed it a cid
        # Otherwise, the event will be created, but the download page
        # will not.
        pages.extend(
            (
                self.summary_page,
                self.download_page,
                self.ami_page,
                self.category_page,
                self.parent_category_page,
            )
        )

        def _handle_existing(err):
            """We need this in two places, so."""
            if err.args[0] == "articleexists":
                # no problem, just move on.
                logger.info(
                    "Page already exists, and forced write was not " "requested! Not writing."
                )
            else:
                raise err

        for pag in pages:
            try:
                # stage 1 - create page
                logger.info("Creating page %s", pag.name)
                pag.write(createonly=createonly)
            except mwclient.errors.APIError as err:
                _handle_existing(err)
            except FedfindNotFoundError:
                # this happens if download page couldn't be created
                # because fedfind release couldn't be found
                logger.warning(
                    "Could not create download page for event %s as fedfind release "
                    "was not found!"
                )

            # stage 2 - update current. this is split so if we hit
            # 'page already exists', we don't skip update_current
            if current and hasattr(pag, "update_current"):
                logger.info("Pointing Current redirect to above page")
                pag.update_current()

        if current:
            try:
                # update CurrentFedoraCompose
                logger.info("Updating CurrentFedoraCompose")
                self.update_current()
            except mwclient.errors.APIError as err:
                _handle_existing(err)

    @classmethod
    def from_page(cls, pageobj):
        """Return the ValidationEvent object for a given ValidationPage
        object.
        """
        return cls(
            pageobj.site, pageobj.release, pageobj.milestone, pageobj.compose, dist=pageobj.dist,
        )


class ComposeEvent(ValidationEvent):
    """An Event that describes a release validation event - that is,
    the testing for a particular nightly, test compose or release
    candidate build.
    """

    def __init__(self, site, release, milestone, compose, dist="Fedora", cid=""):
        super(ComposeEvent, self).__init__(
            site, release, milestone=milestone, compose=compose, dist=dist
        )
        # this is a little hint that gets set via get_validation_event
        # when getting a page or event by cid; it helps us find the
        # fedfind release for the event if the compose is not yet in
        # PDC or synced to stage
        self._cid = cid
        self.shortver = f"{self.milestone} {self.compose}"

    @property
    def category_page(self):
        """The category page for this event. Is a property because
        page instantiation requires a remote trip.
        """
        return wikitcms.listing.ValidationCategory(
            self.site, self.release, self.milestone, dist=self.dist
        )

    @property
    def _current_content(self):
        """The content for the CurrentFedoraCompose template for
        this test event.
        """
        return (
            f"| full = {self.version}\n| release = {self.release}\n"
            f"| milestone = {self.milestone}\n| compose = {self.compose}\n| date =\n"
        )

    @property
    def _pagetype(self):
        """For a ComposeEvent, obviously, ComposePage."""
        return wikitcms.page.ComposePage

    @cached_property
    def creation_date(self):
        """We need this for ordering and determining delta between
        this event and a potential new nightly event, if this is the
        current event. Return creation date of the first result page
        for the event, or "" if it doesn't exist.
        """
        try:
            return self.result_pages[0].creation_date
        except IndexError:
            # event doesn't exist
            return ""

    @property
    def ff_release_images(self):
        """A fedfind release object matching this event, that has
        images. If we can't find one, raise an exception. Here, we
        try getting the release by (release, milestone, compose), but
        if that release has no images - which happens in the specific
        case that we've just created an event for a candidate compose
        which has not yet been synced to stage - and we have the cid
        hint, we try getting a release by cid instead, which should
        find the compose in kojipkgs (a fedfind Production rather than
        Compose).
        """
        rel = self.ff_release
        if rel.all_images:
            return rel

        if self._cid:
            rel = fedfind.release.get_release(cid=self._cid)
        if rel.all_images:
            return rel
        err = f"Could not find fedfind release with images for event {self.version}"
        raise FedfindNotFoundError(err)


class NightlyEvent(ValidationEvent):
    """An Event that describes a release validation event - that is,
    the testing for a particular nightly, test compose or release
    candidate build. Milestone should be 'Rawhide' or 'Branched'.
    Note that a Fedora release number attached to a Rawhide nightly
    compose is an artificial concept that can be considered a Wikitcms
    artifact. Rawhide is a rolling distribution; its nightly composes
    do not really have a release number. What we do when we attach
    a release number to a Rawhide nightly validation test event is
    *declare* that, with our knowledge of Fedora's development cycle,
    we believe the testing of that Rawhide nightly constitutes a part
    of the release validation testing for that future release.
    """

    def __init__(self, site, release, milestone, compose, dist="Fedora"):
        super(NightlyEvent, self).__init__(
            site, release, milestone=milestone, compose=compose, dist=dist
        )
        self.shortver = self.compose
        self.creation_date = compose.split(".")[0]

    @property
    def category_page(self):
        """The category page for this event. Is a property because
        page instantiation requires a remote trip.
        """
        return wikitcms.listing.ValidationCategory(
            self.site, self.release, nightly=True, dist=self.dist
        )

    @property
    def _current_content(self):
        """The content for the CurrentFedoraCompose template for
        this test event.
        """
        return (
            f"| full = {self.version}\n| release = {self.release}\n"
            f"| milestone = {self.milestone}\n| compose =\n| date = {self.compose}\n"
        )

    @property
    def _pagetype(self):
        """For a NightlyEvent, obviously, NightlyPage."""
        return wikitcms.page.NightlyPage


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>
#
"""This file kind of shadows mwclient's listing.py, creating modified
versions of several of its classes. The point is to provide generators
similar to mwclient's own, but which return wikitcms page/category
instances when appropriate, falling through to mwclient instances
otherwise.
"""

import re

import mwclient.listing

import wikitcms.page

# exceptions: the wikitcms getter raises these when it fails rather than
# just returning None, so the generators can use try/except blocks to
# handle both this case and the case (which shouldn't ever happen, but
# just in case) where they're being used on something other than a list
# of pages.


class NoPageWarning(Exception):
    """Exception raised when the tcmswiki getter can't find a matching
    page. Not really an error, should always be handled.
    """

    def __init__(self, page):
        super(NoPageWarning, self).__init__()
        self.page = page

    def __str__(self):
        return f"Could not produce a wikitcms page for: {self.page}"


class PageCheckWarning(Exception):
    """Exception raised when the wikitcms getter finds a matching page,
    but the page name the class generators from the page's various
    attributes doesn't match the page name the getter was given. Should
    usually be handled (and an mwclient Page instance returned instead).
    """

    def __init__(self, frompage, topage):
        super(PageCheckWarning, self).__init__()
        self.frompage = frompage
        self.topage = topage

    def __str__(self):
        return f"Expected page name {self.frompage} does not match source page name {self.topage}"


def _check_page(name, page):
    """A convenience function for _get_tcms sanity check."""
    if page.checkname == name:
        return page
    raise PageCheckWarning(page.checkname, name)


class TcmsGeneratorList(mwclient.listing.GeneratorList):
    """A GeneratorList which returns wikitcms page (and category etc.)
    instances when appropriate. _get_tcms is implemented as a separate
    function so TcmsPageList can use the discovery logic.
    """

    def __next__(self):
        # We can't get the next entry from mwl.List ourselves, try and
        # handle it, then pass it up to our parent if we can't, because
        # parent's next() gets the next entry from mwl.List itself, so
        # in that scenario, one list item gets skipped. Either we
        # entirely clone next() with our own additions, or we let it
        # fire and then override the result if we can. Using nxt._info
        # is bad, but super.next() doesn't return that, and the page
        # instance doesn't expose it any other way. We could just use
        # the name, but if you don't pass info when instantiating a
        # Page, it has to hit the API during init to reconstruct info,
        # and that causes a massive performance hit.
        nxt = super(TcmsGeneratorList, self).__next__()
        try:
            # pylint: disable=protected-access
            return self._get_tcms(nxt.name, nxt._info)
        except (NoPageWarning, PageCheckWarning):
            return nxt

    def _get_tcms_testresults(self, name, info):
        """The main match logic for Test Results pages. Called by
        _get_tcms when it finds a Test Results page.
        """
        nightly_patt = re.compile(
            r"Test Results:Fedora(-\w+)? (\d{1,3}) "
            r"(Rawhide|Nightly|Branched) "
            r"(\d{8,8}(\.n\.\d+)?|\d{4,4} \d{2,2}) "
            r"(.+)$"
        )
        # IoT pages are slightly weird, as IoT composes are production
        # but we treat them as nightlies
        iot_nightly_patt = re.compile(r"Test Results:Fedora-IoT (\d{1,3}) RC (\d{8,8}\.\d+) (.+)$")
        accept_patt = re.compile(
            r"Test Results:Fedora (\d{1,3}) ([^ ]+?) (Rawhide |)Acceptance Test " r"(\d{1,2})$"
        )
        ms_patt = re.compile(r"Test Results:Fedora(-\w+)? (\d{1,3}) ([^ ]+?) ([^ ]+?) (.+)$")

        # Modern standard nightly compose event pages, and F21-era
        # monthly Rawhide/Branched test pages
        match = nightly_patt.match(name)
        if match:
            if match.group(6) == "Summary":
                # we don't really ever need to do anything to existing
                # summary pages, and instantiating one from here is kinda
                # gross, so just fall through
                raise NoPageWarning(name)
            dist = "Fedora"
            if match.group(1):
                dist += match.group(1)
            page = wikitcms.page.NightlyPage(
                self.site,
                release=match.group(2),
                testtype=match.group(6),
                milestone=match.group(3),
                compose=match.group(4),
                info=info,
                dist=dist,
            )
            return _check_page(name, page)

        match = iot_nightly_patt.match(name)
        if match:
            if match.group(3) == "Summary":
                # we don't really ever need to do anything to existing
                # summary pages, and instantiating one from here is kinda
                # gross, so just fall through
                raise NoPageWarning(name)
            page = wikitcms.page.NightlyPage(
                self.site,
                release=match.group(1),
                testtype=match.group(3),
                milestone="RC",
                compose=match.group(2),
                info=info,
                dist="Fedora-IoT",
            )
            return _check_page(name, page)

        match = accept_patt.match(name)
        if match:
            # we don't handle these, yet.
            raise NoPageWarning(name)

        # milestone compose event pages
        match = ms_patt.match(name)
        if match:
            if match.group(5) == "Summary":
                raise NoPageWarning(name)
            dist = "Fedora"
            if match.group(1):
                dist += match.group(1)
            page = wikitcms.page.ComposePage(
                self.site,
                release=match.group(2),
                testtype=match.group(5),
                milestone=match.group(3),
                compose=match.group(4),
                info=info,
                dist=dist,
            )
            return _check_page(name, page)

        # in case we matched nothing...
        raise NoPageWarning(name)

    def _get_tcms_category(self, name, info):
        """The main match logic for Category pages. Call by _get_tcms
        when it finds a Category page.
        """
        cat_patt = re.compile(r"Category:Fedora(-\w+)? (\d{1,3}) " r"(.*?) *?Test Results$")
        tdcat_patt = re.compile(r"Category:Fedora (\d{1,3}) Test Days$")

        # test result categories
        match = cat_patt.match(name)
        if match:
            dist = "Fedora"
            if match.group(1):
                dist += match.group(1)
            if not match.group(3):
                page = ValidationCategory(self.site, match.group(2), info=info, dist=dist)
                return _check_page(name, page)

            if match.group(3) == "Nightly":
                page = ValidationCategory(
                    self.site, match.group(2), nightly=True, info=info, dist=dist
                )
                return _check_page(name, page)

            # otherwise...
            page = ValidationCategory(
                self.site, match.group(2), match.group(3), info=info, dist=dist
            )
            return _check_page(name, page)

        # Test Day categories
        match = tdcat_patt.match(name)
        if match:
            page = TestDayCategory(self.site, match.group(1), info=info)
            return _check_page(name, page)

        # in case we matched nothing...
        raise NoPageWarning(name)

    def _get_tcms_testday(self, name, info):
        """The main match logic for Test Day pages. Called by
        _get_tcms when it finds a Test Day page.
        """
        testday_patt = re.compile(r"Test Day:(\d{4}-\d{2}-\d{2}) *(.*)$")
        # FIXME: There's a few like this, handle 'em sometime
        # testday2_patt = re.compile(u'Test Day:(.+) (\d{4}-\d{2}-\d{2})$')

        match = testday_patt.match(name)
        if match:
            page = wikitcms.page.TestDayPage(self.site, match.group(1), match.group(2), info=info)
            return _check_page(name, page)

        # in case we matched nothing...
        raise NoPageWarning(name)

    def _get_tcms(self, name, info=()):
        """This is where the meat starts: we are going to parse the
        page name and see if it looks like a page we have a class for,
        if so, we will instantiate the class appropriately, check the
        name we get matches the page we're actually running on, and
        return the instance if so. This method quickly sorts pages we
        *might* match into various categories then calls one of the
        specialized methods above to do the rest of the work. At all
        points, if we decide we can't match the page, we raise
        NoPageWarning, which will cause whatever is using us to fall
        through to mwclient.
        """
        if isinstance(name, int):
            # we'll have to do a wiki roundtrip, as we need the text
            # name.
            page = wikitcms.page.Page(self.site, name)
            name = page.name
        name = name.replace("_", " ")
        # quick non-RE checks to see if we'll ever match (and filter
        # out some 'known bad' pages)
        if name.startswith("Test Results:"):
            return self._get_tcms_testresults(name, info)
        if name.startswith("Category:"):
            return self._get_tcms_category(name, info)
        if name.startswith("Test Day:") and not name.endswith("/ru"):
            if not "metadata" in name.lower() and not "rendercheck" in name.lower():
                return self._get_tcms_testday(name, info)

        # otherwise, we'll just bail out here
        raise NoPageWarning(name)


class TcmsPageList(mwclient.listing.PageList, TcmsGeneratorList):
    """A version of PageList which returns wikitcms page (and category
    etc.) objects when appropriate.
    """

    def get(self, name, info=()):
        modname = name
        if self.namespace:
            modname = f"{self.site.namespaces[self.namespace]}:{name}"
        try:
            return self._get_tcms(modname, info)
        except (NoPageWarning, PageCheckWarning):
            return super(TcmsPageList, self).get(name, info)


class TcmsCategory(wikitcms.page.Page, TcmsGeneratorList):
    """A modified category class - just as mwclient's Category class
    inherits from both its Page class and its GeneratorList class,
    acting as both a page and a generator returning the members of
    the category, so this inherits from wikitcms' Page and
    TcmsGeneratorList. You can produce the page contents with pg.Page
    write() method, and you can use it as a generator which returns
    the category's members, as wikitcms class instances if appropriate
    or mwclient class instances otherwise. It works recursively - if
    a member of a ValidationCategory is itself a test category, you'll
    get another ValidationCategory instance. There are sub-classes for
    various particular types of category (Test Days, validation, etc.)
    """

    def __init__(self, site, wikiname, info=None):
        super(TcmsCategory, self).__init__(site, wikiname, info=info)
        TcmsGeneratorList.__init__(self, site, "categorymembers", "cm", gcmtitle=self.name)


class ValidationCategory(TcmsCategory):
    """A category class (inheriting from TcmsCategory) for validation
    test result category pages. If nightly is True, this will be a
    category for test results from Rawhide or Branched nightly builds
    for the given release. Otherwhise, if milestone is passed, this
    will be a category for the given milestone, and if it isn't, it
    will be the top-level category for the given release.
    """

    # no, pylint, this is just the right number of arguments
    # pylint: disable=too-many-arguments
    def __init__(self, site, release, milestone=None, nightly=False, info=None, dist="Fedora"):
        shortdist = dist[7:]
        if nightly is True:
            wikiname = f"Category:{dist} {release} Nightly Test Results"
            self.seedtext = (
                "{{Validation results milestone category"
                f"|release={release}|nightly=true|dist={shortdist}}}}}"
            )

            self.summary = (
                "Relval bot-created validation result category "
                f"page for {dist} {release} nightly results"
            )
        elif milestone:
            wikiname = f"Category:{dist} {release} {milestone} Test Results"
            self.seedtext = (
                f"{{{{Validation results milestone category|release={release}"
                f"|milestone={milestone}|dist={shortdist}}}}}"
            )
            self.summary = (
                f"Relval bot-created validation result category page for {dist} "
                f"{release} {milestone}"
            )
        else:
            wikiname = f"Category:{dist} {release} Test Results"
            self.seedtext = (
                "{{Validation results milestone category" f"|release={release}|dist={shortdist}}}}}"
            )
            self.summary = (
                "Relval bot-created validation result category page for " f"{dist} {release}"
            )

        super(ValidationCategory, self).__init__(site, wikiname, info=info)


class TestDayCategory(TcmsCategory):
    """A category class (inheriting from TcmsCategory) for Test Day
    category pages.
    """

    def __init__(self, site, release, info=None):
        wikiname = f"Category:Fedora {str(release)} Test Days"
        self.seedtext = (
            f"This category contains all the Fedora {str(release)} [[QA/Test_Days|Test "
            "Day]] pages. A calendar of the Test Days can be found ["
            "https://apps.fedoraproject.org/calendar/list/QA/?subject=Test+Day"
            " here].\n\n[[Category:Test Days]]"
        )
        self.summary = "Created page (via wikitcms)"
        super(TestDayCategory, self).__init__(site, wikiname, info=info)


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>
#
"""This file contains helper functions that don't strictly belong in
any particular class or even in another file but outside of a class."""

import re
from decimal import Decimal

import fedfind.helpers
import fedfind.release

MILESTONE_PAIRS = (
    ("Rawhide", "100"),
    # We called Branched nightly pages 'Nightly' in F21 cycle.
    ("Nightly", "149"),
    ("Branched", "150"),
    ("Pre-Alpha", "175"),
    ("Alpha", "200"),
    ("Pre-Beta", "375"),
    ("Beta", "400"),
    ("Preview", "600"),
    ("Pre-Final", "775"),
    ("Final", "800"),
    ("RC", "800"),
    ("Postrelease", "900"),
)

COMPOSE_PAIRS = (
    # Some F12 crap
    ("PreBeta", "100"),
    ("TC", "200"),
    # and this.
    ("Pre-RC", "300"),
    # The extra digit here is a dumb way to make sure RC1 sorts
    # later than TC10 - otherwise you get 20010 vs. 6001, and 20010
    # wins. It might be better to treat the 'TC' / 'RC' as a separate
    # element in triplet_sort, but that's a bit harder and makes its
    # name a lie...
    ("RC", "6000"),
)


def fedora_release_sort(string):
    """Fed a string that looks something like a Fedora pre-release /
    compose version, this will output a modified version of the string
    which should sort correctly against others like it. Handles stuff
    like 'Preview' coming before 'Final', and 'TC' coming before 'RC'.
    'Alpha', 'Beta' and 'Final' sort correctly by pure chance, but we
    handle them here anyway to make sure. wikitcms ValidationEvent and
    ValidationPage objects have a 'sortname' property you can use
    instead of calling this directly. Milestones have a letter so they
    sort after nightlies (nightlies are usually going to be earlier).
    NOTE: can only sort compose event versions, really. With this
    function, '22 Alpha TC1' > '22 Alpha'.
    """
    # Some MILESTONES are substrings of COMPOSES so we do Cs first
    for (orig, repl) in COMPOSE_PAIRS + MILESTONE_PAIRS:
        string = string.replace(orig, repl)
    return string


def triplet_sort(release, milestone, compose):
    """Just like fedora_release_sort, but requires you to pass the
    now-'standard' release, milestone, compose triplet of inputs.
    This is a better way in most cases as you're going to have more
    certainty about instantiating wikitcms/fedfind objects from it,
    plus we can handle things like '23' being higher than '23 Beta'
    or '23 Final TC1'. Expects the inputs to be strings. The elements
    in the output tuple will be ints if possible as this gives a
    better sort, but may be strings if we missed something; you're not
    meant to *do* anything with the tuple but compare it to another
    similar tuple.
    """
    for (orig, repl) in MILESTONE_PAIRS:
        milestone = milestone.replace(orig, repl)
        if not milestone:
            # ('23', 'Final', '') == ('23', '', '')
            milestone = "800"
    for (orig, repl) in COMPOSE_PAIRS:
        compose = compose.replace(orig, repl)
        if not compose:
            compose = "999"
    # We want to get numerical sorts if we possibly can, so e.g.
    # TC10 (becomes 20010) > TC9 (becomes 2009). But just in case
    # we get passed a character we don't substitute to a digit,
    # check first.
    release = str(release)
    if release.isdigit():
        release = int(release)
    if milestone.isdigit():
        milestone = int(milestone)
    if compose.isdigit():
        compose = int(compose)
    else:
        # this is a bit magic but handles "TC1.1" (old-school),
        # "1.1" (new-school candidates), and "20160314.n.0" (new-
        # school nightlies)
        (comp, respin) = (compose.split(".")[0], compose.split(".")[-1])
        if comp.isdigit() and respin.isdigit():
            compose = Decimal(".".join((comp, respin)))
    return (release, milestone, compose)


def triplet_unsort(release, milestone, compose):
    """Reverse of triplet_sort."""
    (release, milestone, compose) = (str(release), str(milestone), str(compose))
    for (repl, orig) in MILESTONE_PAIRS:
        milestone = milestone.replace(orig, repl)
        if milestone == "800":
            milestone = "Final"
    # We don't want to do the replace here if the compose is a date,
    # because any of the values might happen to be in the date. This
    # is a horrible hack that should be OK (until 2100, at least).
    if fedfind.helpers.date_check(compose, fail_raise=False):
        pass
    elif fedfind.helpers.date_check(compose.split(".")[0], fail_raise=False):
        # this is slightly magic but should do for now
        (comp, respin) = compose.split(".")
        compose = f"{comp}.n.{respin}"
    else:
        for (repl, orig) in COMPOSE_PAIRS:
            compose = compose.replace(orig, repl)
            if compose == "999":
                compose = ""
    return (release, milestone, compose)


def rreplace(string, old, new, occurrence):
    """A version of the str.replace() method which works from the right.
    Taken from https://stackoverflow.com/questions/2556108/
    """
    elems = string.rsplit(old, occurrence)
    return new.join(elems)


def normalize(text):
    """Lower case and replace ' ' with '_' so I don't have to
    keep retyping it.
    """
    return text.lower().replace(" ", "_")


def find_bugs(text):
    """Find RH bug references in a given chunk of text. More than one
    method does this, so we'll put the logic here and they can share
    it. Copes with [[rhbug:(ID)]] links, {{bz|(ID)}} templates and
    URLs that look like Bugzilla. Doesn't handle aliases (only matches
    numeric IDs 6 or 7 digits long). Returns a set (so bugs that occur
    multiple times in the text will only appear once in the output).
    """
    bugs = set()
    bzpatt = re.compile(
        r"({{bz *\| *|" r"\[\[rhbug *: *|" r"bugzilla\.redhat\.com/show_bug\.cgi\?id=)" r"(\d{6,7})"
    )
    matches = bzpatt.finditer(text)
    for match in matches:
        bugs.add(match.group(2))
    # Filter out bug IDs usually used as examples
    for bug in ("12345", "123456", "54321", "654321", "234567", "1234567", "7654321"):
        bugs.discard(bug)
    return bugs


def cid_to_event(cid):
    """Given a Pungi 4 compose ID, figure out the appropriate wikitcms
    (release, milestone, compose) triplet. Guesses the appropriate
    release number to assign to the event for Rawhide composes. To
    do the opposite, you can use the ff_release property of Validation
    Event instances.
    """
    parsed = fedfind.helpers.parse_cid(cid, dic=True)
    dist = parsed["short"]
    release = parsed["version"].lower()
    vertype = parsed["version_type"]
    date = parsed["date"]
    typ = parsed["compose_type"]
    respin = parsed["respin"]
    if dist not in ("Fedora", "Fedora-Modular", "Fedora-IoT"):
        # we only have validation events for these composes ATM -
        # not e.g. FedoraRespin or Fedora-Atomic ones
        raise ValueError(f"No validation events for {dist} composes!")
    if vertype != "ga":
        # this ensures we don't create events for updates/u-t composes
        raise ValueError("No validation events for updates composes!")
    if typ == "production":
        # we need to get the label and parse that
        ffrel = fedfind.release.get_release(cid=cid)
        if not ffrel.label:
            raise ValueError(
                f"{cid} looks like a production compose, but found "
                "no label! Cannot determine event."
            )
        (milestone, compose) = ffrel.label.rsplit("-", 1)
        return (dist, release, milestone, compose)

    if not typ == "nightly":
        # we don't have validation events for any compose types other
        # than 'production' and 'nightly'
        raise ValueError("cid_to_event(): cannot guess event for 'test' compose")

    # nightlies
    if release == "rawhide":
        # the release number for new Rawhide events should always be
        # 1 above the highest current 'real' release (branched or
        # stable). FIXME: this will not do the right thing for *old*
        # composes, not sure if we can fix that sensibly.
        release = str(fedfind.helpers.get_current_release(branched=True) + 1)
        milestone = "Rawhide"
    else:
        milestone = "Branched"
    compose = f"{date}.n.{respin}"
    return (dist, release, milestone, compose)


def validate_dist(dist):
    """Just a quick check that dist names match the expected form."""
    distpatt = re.compile(r"^Fedora(-\w+)?$")
    if distpatt.match(dist):
        return dist
    raise ValueError("%s is not a valid dist!" % dist)


# vim: set textwidth=100 ts=8 et sw=4:



# Copyright (C) 2020 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# these are all kinda inappropriate for pytest patterns
# pylint: disable=no-init, protected-access, no-self-use, unused-argument
# pylint: disable=invalid-name

"""Tests for listing.py."""

from unittest import mock

import mwclient.image
import mwclient.page
import pytest

import wikitcms.listing
import wikitcms.page
import wikitcms.release
import wikitcms.wiki


@pytest.mark.usefixtures("fakemwp")
class TestListing:
    """Tests for the generator magic in listing.py. This stuff is all
    a bit odd, and because of that, the testing is a bit indirect. A
    lot of the tests don't go directly through anything defined in
    listing.py but through Wiki.pages, because ultimately backing that
    is the most important thing listing.py does: Wiki.pages is an
    instance of listing's TcmsPageList, which is a subclass of
    TcmsGeneratorList, where most of the magic lives. Aside from that
    we test the category stuff via the 'real' category classes that
    use it.
    """

    # the "fjajah" is just to make sure we're running offline; if I
    # screw up and add a test that hits the network it"ll cause the
    # tests to hang/fail instead of succeeding a bit slower
    site = wikitcms.wiki.Wiki("fjajah", do_init=False, force_login=False)

    @mock.patch("mwclient.image.Image.__init__", return_value=None)
    def test_pages_nomatch(self, fakeimage):
        """If we don't hit any of the special Wikitcms classes, we
        should return whatever mwclient would return - a Page, Image,
        or Category. We also should not hit re.compile for Page or
        Image, we should catch them with early checks.
        """
        with mock.patch("re.compile"):
            assert isinstance(self.site.pages["Somepage"], mwclient.page.Page)
            assert isinstance(self.site.pages["Image:Someimage"], mwclient.image.Image)
        # we do handle some categories, so we can't mock re.compile
        assert isinstance(self.site.pages["Category:Something"], mwclient.listing.Category)

    @mock.patch("re.compile")
    def test_pages_knownbad_early(self, fakecompile):
        """Some more cases where we should just fall through, and do
        it before hitting re.compile. Mocking it out effectively tests
        that: if we hit it, things will blow up because it's not doing
        its job because it's a MagicMock.
        """
        # we don't handle the translated pages, for test days where
        # translation happened
        assert isinstance(self.site.pages["Test_Day:2010-09-29_Radeon/ru"], mwclient.page.Page)
        # the metadata pages are special stuff for the testdays webapp
        pg = self.site.pages["Test Day:2019-09-23 F31 Upgrade Test Day TestdayApp Metadata"]
        assert isinstance(pg, mwclient.page.Page)
        # these were subpages of the Test Day page proper where we
        # put output from a specific test
        pg = self.site.pages["Test Day:2011-02-22 Nouveau/rendercheck/red"]
        assert isinstance(pg, mwclient.page.Page)

    def test_pages_knownbad_late(self):
        """Some more cases where we should just fall through, but
        later in the process, because these pages look kinda like we
        should handle them but we don't.
        """
        # a very early validation page style we don't handle yet
        pg = self.site.pages["Test Results:Fedora 13 Pre-RC Acceptance Test 1"]
        assert isinstance(pg, mwclient.page.Page)

    def test_pages_nightly(self):
        """Nightly validation pages."""
        pg = self.site.pages["Test_Results:Fedora_32_Branched_20200322.n.0_Installation"]
        assert isinstance(pg, wikitcms.page.NightlyPage)
        assert pg.release == "32"
        assert pg.testtype == "Installation"
        assert pg.milestone == "Branched"
        assert pg.compose == "20200322.n.0"
        assert pg.dist == "Fedora"
        # yeah, the _ vs. space thing is intentional, we ought to
        # handle both, may as well check it as we go
        pg = self.site.pages["Test Results:Fedora 32 Rawhide 20191030.n.1 Base"]
        assert isinstance(pg, wikitcms.page.NightlyPage)
        assert pg.testtype == "Base"
        assert pg.milestone == "Rawhide"
        # modular
        pg = self.site.pages["Test Results:Fedora-Modular 27 Branched 20171123.n.1 Server"]
        assert isinstance(pg, wikitcms.page.NightlyPage)
        assert pg.dist == "Fedora-Modular"
        # IoT
        pg = self.site.pages["Test Results:Fedora-IoT 33 RC 20200513.0 General"]
        assert isinstance(pg, wikitcms.page.NightlyPage)
        assert pg.dist == "Fedora-IoT"
        # summary should fall through to an mwclient Page
        pg = self.site.pages["Test_Results:Fedora_32_Branched_20200322.n.0_Summary"]
        assert not isinstance(pg, wikitcms.page.Page)
        assert isinstance(pg, mwclient.page.Page)

    def test_pages_f21_monthly(self):
        """The old monthly pages we had around F21."""
        pg = self.site.pages["Test Results:Fedora 21 Nightly 2014 08 Base"]
        assert isinstance(pg, wikitcms.page.NightlyPage)
        assert pg.release == "21"
        assert pg.milestone == "Nightly"
        assert pg.compose == "2014 08"
        assert pg.testtype == "Base"
        pg = self.site.pages["Test Results:Fedora 21 Rawhide 2014 04 Installation"]
        assert isinstance(pg, wikitcms.page.NightlyPage)
        assert pg.milestone == "Rawhide"
        assert pg.testtype == "Installation"

    def test_pages_milestone_f12(self):
        """Milestone validation pages, weird cases from F12."""
        pg = self.site.pages["Test Results:Fedora 12 Final Pre-RC Install"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "12"
        assert pg.milestone == "Final"
        assert pg.compose == "Pre-RC"
        pg = self.site.pages["Test Results:Fedora 12 Alpha TC Install"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.milestone == "Alpha"
        assert pg.compose == "TC"
        pg = self.site.pages["Test Results:Fedora 12 Alpha TCRegression Install"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.milestone == "Alpha"
        assert pg.compose == "TCRegression"
        pg = self.site.pages["Test Results:Fedora 12 Beta PreBeta Install"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.milestone == "Beta"
        assert pg.compose == "PreBeta"

    def test_pages_milestone_tcrc(self):
        """Milestone validation pages, the pre-Pungi 4 "TCx / RCx"
        style.
        """
        # also testing the old "Install" (not "Installation") testtype
        pg = self.site.pages["Test Results:Fedora 20 Alpha RC1 Install"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "20"
        assert pg.milestone == "Alpha"
        assert pg.compose == "RC1"
        assert pg.testtype == "Install"
        assert pg.dist == "Fedora"
        pg = self.site.pages["Test Results:Fedora 22 Beta TC1 Installation"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "22"
        assert pg.milestone == "Beta"
        assert pg.compose == "TC1"
        assert pg.testtype == "Installation"

    def test_pages_milestone_pungi4(self):
        """Milestone validation pages, current (2020-04) Pungi 4
        style.
        """
        pg = self.site.pages["Test Results:Fedora 24 Alpha 1.1 Cloud"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "24"
        assert pg.milestone == "Alpha"
        assert pg.compose == "1.1"
        assert pg.testtype == "Cloud"
        assert pg.dist == "Fedora"
        pg = self.site.pages["Test Results:Fedora 26 Beta 1.3 Security Lab"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "26"
        assert pg.milestone == "Beta"
        assert pg.compose == "1.3"
        assert pg.testtype == "Security Lab"
        pg = self.site.pages["Test Results:Fedora 31 RC 1.9 Cloud"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "31"
        assert pg.milestone == "RC"
        assert pg.compose == "1.9"
        assert pg.testtype == "Cloud"
        # modular
        pg = self.site.pages["Test Results:Fedora-Modular 27 Beta 1.5 Installation"]
        assert isinstance(pg, wikitcms.page.ComposePage)
        assert pg.release == "27"
        assert pg.milestone == "Beta"
        assert pg.compose == "1.5"
        assert pg.testtype == "Installation"
        assert pg.dist == "Fedora-Modular"
        # again, summary should fall through to an mwclient Page
        pg = self.site.pages["Test Results:Fedora 31 RC 1.9 Summary"]
        assert not isinstance(pg, wikitcms.page.Page)
        assert isinstance(pg, mwclient.page.Page)

    def test_pages_testday(self):
        """Test Day pages."""
        pg = self.site.pages["Test Day:2020-04-08 Fedora 32 IoT Edition"]
        assert isinstance(pg, wikitcms.page.TestDayPage)
        assert pg.date == "2020-04-08"
        assert pg.subject == "Fedora 32 IoT Edition"
        # why this exists I don't know, but it does
        pg = self.site.pages["Test Day:2012-03-14"]
        assert isinstance(pg, wikitcms.page.TestDayPage)
        assert pg.date == "2012-03-14"
        assert pg.subject == ""

    # we're just mocking this so we can count the calls
    @mock.patch("wikitcms.listing.PageCheckWarning.__init__", autospec=True, return_value=None)
    def test_pages_checkname(self, fakepagecheck):
        """There's a mechanism where we refuse to return a Wikitcms-y
        instance if its expected name (which all Wikitcms Page classes
        are required to define, as 'checkname') does not match the
        actual page name we got the instance from. This checks that
        still works.
        """
        # this should match the testday_patt regex, but the extra
        # spaces should mean the checkname won't match
        pg = self.site.pages["Test Day:2020-04-08  Foobar"]
        assert isinstance(pg, mwclient.page.Page)
        # just the above isn't sufficient, as our classes are subs of
        # mwclient Page
        assert not isinstance(pg, wikitcms.page.TestDayPage)
        # to be super sure this worked as intended, check we hit the
        # exception's __init__
        assert fakepagecheck.call_count == 1

    def test_pages_validation_category(self):
        """Validation category pages."""
        # top level
        pg = self.site.pages["Category:Fedora 32 Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == "{{Validation results milestone category|release=32|dist=}}"
        assert pg.summary == "Relval bot-created validation result category page for Fedora 32"
        # top level, modular
        pg = self.site.pages["Category:Fedora-Modular 27 Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == "{{Validation results milestone category|release=27|dist=Modular}}"
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora-Modular 27"
        )
        # top level, IoT
        pg = self.site.pages["Category:Fedora-IoT 33 Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == "{{Validation results milestone category|release=33|dist=IoT}}"
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora-IoT 33"
        )
        # milestone level
        pg = self.site.pages["Category:Fedora 32 Beta Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == (
            "{{Validation results milestone category|release=32|milestone=Beta|dist=}}"
        )
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora 32 Beta"
        )
        # milestone level, modular
        pg = self.site.pages["Category:Fedora-Modular 27 Beta Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == (
            "{{Validation results milestone category|release=27|milestone=Beta|dist=Modular}}"
        )
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora-Modular 27 Beta"
        )
        # nightly level
        pg = self.site.pages["Category:Fedora 32 Nightly Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == (
            "{{Validation results milestone category|release=32|nightly=true|dist=}}"
        )
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora 32 nightly results"
        )
        # nightly level, modular
        pg = self.site.pages["Category:Fedora-Modular 27 Nightly Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == (
            "{{Validation results milestone category|release=27|nightly=true|dist=Modular}}"
        )
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora-"
            "Modular 27 nightly results"
        )
        # nightly level, IoT
        pg = self.site.pages["Category:Fedora-IoT 33 Nightly Test Results"]
        assert isinstance(pg, wikitcms.listing.ValidationCategory)
        assert pg.seedtext == (
            "{{Validation results milestone category|release=33|nightly=true|dist=IoT}}"
        )
        assert pg.summary == (
            "Relval bot-created validation result category page for Fedora-"
            "IoT 33 nightly results"
        )

    @mock.patch("mwclient.listing.GeneratorList.__next__")
    def test_pages_testday_category(self, fakenext):
        """Test Day category pages."""
        pg = self.site.pages["Category:Fedora 32 Test Days"]
        assert isinstance(pg, wikitcms.listing.TestDayCategory)
        assert pg.seedtext == (
            "This category contains all the Fedora 32 [[QA/Test_Days|Test "
            "Day]] pages. A calendar of the Test Days can be found ["
            "https://apps.fedoraproject.org/calendar/list/QA/?subject=Test+Day"
            " here].\n\n[[Category:Test Days]]"
        )
        assert pg.summary == "Created page (via wikitcms)"
        # test the iterator capability
        tdp = mwclient.page.Page(self.site, "Test Day:2019-12-09 Kernel 5.4 Test Week")
        fakenext.return_value = tdp
        assert isinstance(next(pg), wikitcms.page.TestDayPage)
        # iterator when page isn't one of ours
        fakenext.return_value = mwclient.page.Page(self.site, "Somepage")
        nextpg = next(pg)
        assert not isinstance(nextpg, wikitcms.page.Page)
        assert isinstance(nextpg, mwclient.page.Page)

    def test_pages_int_name(self):
        """We handle getting an integer as the page 'name' - which
        IIRC retrieves a page by some sort of unique ID - by doing a
        wiki roundtrip to get the real string name.
        """
        pg = self.site.pages[80736]
        assert isinstance(pg, wikitcms.page.TestDayPage)

    def test_allresults(self):
        """This is a test using the allresults generator, because that
        exercises the namespace handling in TcmsPageList.
        """
        # this would usually be set up by a remote trip in
        # site.__init__ but obviously not with do_init=False, so we
        # have to just set it up manually
        self.site.namespaces[116] = "Test Results"
        pg = self.site.allresults()["Fedora_32_Branched_20200322.n.0_Installation"]
        assert isinstance(pg, wikitcms.page.NightlyPage)

    def test_alltestdays(self):
        """This is a test using the alltestdays generator. We may as
        well test it here too as it's just the same as allresults.
        """
        # this would usually be set up by a remote trip in
        # site.__init__ but obviously not with do_init=False, so we
        # have to just set it up manually
        self.site.namespaces[114] = "Test Day"
        pg = self.site.alltestdays()["2019-12-09 Kernel 5.4 Test Week"]
        assert isinstance(pg, wikitcms.page.TestDayPage)

    def test_nopagewarning(self):
        """Test for the NoPageWarning exception."""
        exc = wikitcms.listing.NoPageWarning("QA:Somepage")
        assert str(exc) == "Could not produce a wikitcms page for: QA:Somepage"

    def test_pagecheckwarning(self):
        """Test for the PageCheckWarning exception."""
        exc = wikitcms.listing.PageCheckWarning("QA:Somepage", "QA:Someotherpage")
        assert str(exc) == (
            "Expected page name QA:Somepage does not match source " "page name QA:Someotherpage"
        )



# Copyright (C) 2014 Red Hat
#
# This file is part of python-wikitcms.
#
# python-wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

"""Classes that describe different types of pages we are interested
in, and attributes of pages like test results and test cases, are
defined in this file.
"""

from collections import OrderedDict
import datetime
import logging
import re
import time

from cached_property import cached_property
import fedfind.helpers
import fedfind.release
from mwclient import errors as mwe
from mwclient import page as mwp
import pytz

from wikitcms.exceptions import NoPageError, NotFoundError, TooManyError
import wikitcms.helpers
import wikitcms.result

logger = logging.getLogger(__name__)


class Page(mwp.Page):
    """Parent class for all page classes. Can be instantiated directly
    if you just want to take advantage of the convenience methods like
    sections() and save(). Available attributes: seedtext, summary.
    Note 'name' is defined by mwp.Page's __init__.
    """

    def __init__(self, site, wikiname, info=None, extra_properties=None):
        super(Page, self).__init__(site, wikiname, info, extra_properties)
        # Used for sanity check by the page generator
        self.checkname = wikiname
        self._sections = None
        self.results_separators = list()

    @property
    def sections(self):
        """A list of the page's sections. Each section is represented
        by a dict whose values provide various attributes of the
        section. Returns an empty list for non-existent page (or any
        other API error). Cached, cache cleared on each page save.
        """
        # None == not yet retrieved or cache expired. [] == retrieved,
        # but page is empty or something.
        if self._sections is None:
            try:
                apiout = self.site.api("parse", page=self.name, prop="sections")
                self._sections = apiout["parse"]["sections"]
            except mwe.APIError:
                self._sections = []
        return self._sections

    @property
    def results_wikitext(self):
        """Returns a string containing the wikitext for the page's
        results section. Will be empty if no results are found. Relies
        on the layout for result pages remaining consistent. Class
        must override definition of self.results_separators or else
        this will always return an empty string.
        """
        pagetext = self.text()
        comment = re.compile("<!--.*?-->", re.S)
        pos = -1
        for sep in self.results_separators:
            pos = pagetext.find(sep)
            if pos > -1:
                break

        if pos == -1:
            return ""
        text = pagetext[pos:]
        text = comment.sub("", text)
        return text

    @cached_property
    def creation_date(self):
        """Date the page was created. Used for sorting and seeing how
        long it's been since the last event, when creating new events.
        """
        revs = self.revisions(limit=1, dir="newer", prop="timestamp")
        try:
            origrev = next(revs)
        except StopIteration:
            # page doesn't exist
            return ""
        return time.strftime("%Y%m%d", origrev["timestamp"])

    def write(self, createonly=True):
        """Create a page with its default content and summary. mwclient
        exception will be raised on any page write failure.
        """
        seedtext = getattr(self, "seedtext", None)
        summary = getattr(self, "summary", None)
        if seedtext is None or summary is None:
            raise ValueError("wikitcms.Page.write(): both seedtext and summary needed!")
        self.save(seedtext, summary, createonly=createonly)

    # we're using *args, **kwargs. cool your jets, pylint.
    # pylint: disable=signature-differs
    def save(self, *args, **kwargs):
        """Same as the original, but will retry once on fail. If you
        already retrieved the current text, you can pass it in as
        oldtext, and we will check to see if oldtext and text are the
        same. If they are, we return a dict with the key nochange set
        to an empty string - this saves a needless extra remote round
        trip. Of course you could do this in the caller instead, it's
        just a convenience. Also clears the page sections cache.
        """
        if "oldtext" in kwargs and args[0] == kwargs["oldtext"]:
            return dict(nochange="")

        if "oldtext" in kwargs:
            # avoid mwclient save() warning about unknown kwarg
            del kwargs["oldtext"]

        try:
            ret = super(Page, self).save(*args, **kwargs)
        except mwe.EditError as err:
            logger.warning("Page %s edit failed! Trying again in 15 seconds", self.name)
            logger.debug("Error was: %s", err)
            time.sleep(15)
            ret = super(Page, self).save(*args, **kwargs)
        # clear the caches
        self._sections = None
        return ret


class ValidationPage(Page):
    """A parent class for different types of release validation event
    pages, containing common properties and methods. Required
    attributes: version, shortver, seedtext. "dist" lets us create
    multiple event streams for different compose streams: it defaults
    to "Fedora", which gives us page names for the main stream of
    composes and Fedora releases, but can be set to e.g. "Fedora IoT"
    for a validation page for an IoT compose. The provided string is
    simply included in page names, category names and so on.
    """

    # I like this number, pylint
    # pylint: disable=too-many-instance-attributes, too-many-arguments
    def __init__(self, site, release, testtype, milestone="", compose="", info=None, dist="Fedora"):
        self.release = release
        self.milestone = str(milestone)
        try:
            self.compose = fedfind.helpers.date_check(compose, fail_raise=True, out="str")
        except ValueError:
            self.compose = str(compose)
        self.version = f"{self.release} {self.milestone} {self.compose}"
        self.testtype = testtype
        self.dist = dist
        # dist without "Fedora-", used in various wiki templates;
        # for the main "Fedora" composes, this will be ""
        self.shortdist = dist[7:]

        # Wiki name the page should have, according to the naming
        # convention.
        wikiname = f"Test Results:{dist} {self.version} {self.testtype}"
        super(ValidationPage, self).__init__(site, wikiname, info)

        # Edit summary to be used for clean page creation.
        self.summary = (
            f"Relval bot-created {testtype} validation results page for {dist} " f"{self.version}"
        )
        self.results_separators = (
            "Test Matri",
            "Test Areas",
            "An unsupported test or configuration.  No testing is required.",
        )
        # Sorting helpers. sortname is a string, sorttuple is a
        # 4-tuple. sorttuple is more reliable. See the function docs.
        self.sortname = wikitcms.helpers.fedora_release_sort(
            " ".join((self.version, self.testtype))
        )
        self.sorttuple = wikitcms.helpers.triplet_sort(
            self.release, self.milestone, self.compose
        ) + (self.testtype,)

    @property
    def results_sections(self):
        """A list of the sections in the page which (most likely)
        contain test results. Takes all the sections in the page,
        finds the one one which looks like the first "test results"
        section and returns that section and those that follow it - or
        returns all sections after the Key section, if it can't find
        one which looks like the first results section.
        """
        secs = self.sections
        if not secs:
            # empty page or some other malarkey
            return secs
        first = None
        for i, sec in enumerate(secs):
            if "Test Matri" in sec["line"] or "Test Areas" in sec["line"]:
                first = i
                break
            if "Key" in sec["line"]:
                first = i + 1
        return secs[first:]

    def get_resultrows(self, statuses=None, transferred=True):
        """Returns the result.ResultRow objects representing all the
        page's table rows containing test results.
        """
        sections = self.results_sections
        if not sections:
            return list()
        rows = list()
        pagetext = self.text()
        comment = re.compile("<!--.*?-->", re.S)
        for i, sec in enumerate(sections):
            try:
                nextsec = sections[i + 1]
            except IndexError:
                nextsec = None
            section = sec["line"]
            secid = sec["index"]
            if nextsec:
                sectext = pagetext[sec["byteoffset"] : nextsec["byteoffset"]]
            else:
                sectext = pagetext[sec["byteoffset"] :]
            # strip comments
            sectext = comment.sub("", sectext)
            newrows = wikitcms.result.find_resultrows(
                sectext, section, secid, statuses, transferred
            )
            rows.extend(newrows)
        return rows

    def find_resultrow(self, testcase="", section="", testname="", env=""):
        """Return exactly one result row with the desired attributes,
        or raise an exception (if more or less than one row is found).
        The Installation page contains some rows in the same section
        with the same testcase and testname, but each row provides
        a different set of envs, so these can be uniquely identified
        by specifying the desired env.
        """
        rows = self.get_resultrows()
        if not rows:
            raise NoPageError("Page does not exist or has no result rows.")

        # Find the right row
        nrml = wikitcms.helpers.normalize
        rows = [
            r for r in rows if nrml(testcase) in nrml(r.testcase) or nrml(testcase) in nrml(r.name)
        ]
        if len(rows) > 1 and section:
            rows = [r for r in rows if nrml(section) in nrml(r.section)]
        if len(rows) > 1 and testname:
            rows = [r for r in rows if nrml(testname) in nrml(r.name)]
        if len(rows) > 1 and env:
            # the way this match is done must be kept in line with the
            # corresponding match in add_results, below
            rows = [r for r in rows if nrml(env) in [renv.lower() for renv in r.results.keys()]]
        # try a more precise name match - e.g. "upgrade_dnf" vs.
        # "upgrade_dnf_encrypted"
        if len(rows) > 1:
            rows = [
                r
                for r in rows
                if nrml(testcase) == nrml(r.testcase)
                or nrml(testcase) == nrml(r.name)
                or nrml(testname) == nrml(r.name)
            ]
        if not rows:
            raise NotFoundError("Specified row cannot be found.")
        if len(rows) > 1:
            raise TooManyError("More than one matching row found.")
        return rows[0]

    def update_current(self):
        """Make the Current convenience redirect page on the wiki for
        the given test type point to this page.
        """
        short = self.shortdist
        if short:
            short = f" {short}"
        curr = self.site.pages[f"Test Results:Current{short} {self.testtype} Test"]
        curr.save(f"#REDIRECT [[{self.name}]]", "relval: update to current event", createonly=None)

    def add_results(self, resultsdict, allowdupe=False):
        """Adds multiple results to the page. Passed a dict whose
        keys are ResultRow() instances and whose values are iterables
        of (env, Result()) 2-tuples. Returns a list, which will be
        empty unless allowdupe is False and any of the results is a
        'dupe' - i.e. the given test and environment already have a
        result from the user. The return list contains a 3-tuple of
        (row, env, result) for each dupe.
        """
        # look, pylint, this is complicated stuff. calm yoself.
        # pylint: disable=too-many-locals, too-many-branches, too-many-statements

        # We need to sort the dict in a particular way: by the section
        # ID of each row, in reverse order. This is so when we edit
        # the page, we effectively do so backwards, and the byte
        # offsets we use to find each section don't get thrown off
        # along the way (we don't edit section 1 before section 3 and
        # thus not quite slice the text correctly when we look for
        # section 3).
        resultsdict = OrderedDict(
            sorted(resultsdict.items(), key=lambda x: int(x[0].secid), reverse=True)
        )
        nonetext = wikitcms.result.Result().result_template
        dupes = list()
        # this tracks rows that we don't wind up touching, so they
        # can be left out of the summary text
        notrows = list()
        newtext = oldtext = self.text()
        for (row, results) in resultsdict.items():
            # It's possible that we have rows with identical text in
            # different page sections; this is why 'secid' is an attr
            # of ResultRows. To make sure we edit the correct row,
            # we'll slice the text at the byteoffset of the row's
            # section. We only do one replacement, so we don't need
            # to bother finding the *end* of the section.
            # We could just edit the page section-by-section, but that
            # involves doing one remote roundtrip per section.
            secoff = [sec["byteoffset"] for sec in self.sections if sec["index"] == row.secid][0]
            if secoff:
                sectext = newtext[secoff:]
            else:
                sectext = newtext
            oldrow = row.origtext
            cells = oldrow.split("\n|")

            for (env, result) in results:
                if not env in row.results:
                    # the env passed wasn't precisely one of the row's
                    # envs. let's see if we can make a safe guess. If
                    # there's only one env, it's easy...
                    if len(row.results) == 1:
                        env = list(row.results.keys())[0]
                    else:
                        # ...if not, we'll see if the passed env is
                        # a substring of only one of the envs, case-
                        # insensitively.
                        cands = [cand for cand in row.results.keys() if env.lower() in cand.lower()]
                        if len(cands) == 1:
                            env = cands[0]
                        else:
                            # LOG: bad env
                            continue
                if not allowdupe:
                    dupe = [r for r in row.results[env] if r.user == result.user]
                    if dupe:
                        dupes.append((row, env, result))
                        continue
                restext = result.result_template
                rescell = cells[row.columns.index(env)]
                if nonetext in rescell:
                    rescell = rescell.replace(nonetext, restext)
                elif "\n" in rescell:
                    rescell = rescell.replace("\n", restext + "\n")
                else:
                    rescell = rescell + restext
                cells[row.columns.index(env)] = rescell

            newrow = "\n|".join(cells)
            if newrow == oldrow:
                # All dupes, or something.
                notrows.append(row)
                continue
            sectext = sectext.replace(oldrow, newrow, 1)
            if secoff:
                newtext = newtext[:secoff] + sectext
            else:
                newtext = sectext

        touchedrows = [row for row in resultsdict.keys() if row not in notrows]
        if len(touchedrows) > 3:
            testtext = ", ".join(row.name for row in touchedrows[:3])
            testtext = f"{testtext}..."
        else:
            testtext = ", ".join(row.name for row in touchedrows)
        summary = f"Result(s) for test(s): {testtext} filed via relval"
        self.save(newtext, summary, oldtext=oldtext, createonly=None)
        return dupes

    def add_result(self, result, row, env, allowdupe=False):
        """Adds a result to the page. Must be passed a Result(), the
        result.ResultRow() object representing the row into which a
        result will be added, and the name of the environment for
        which the result is to be reported. Works by replacing the
        first instance of the row's text encountered in the page or
        page section. Expected to be used together with get_resultrows
        which provides the ResultRow() objects.
        """
        resdict = dict()
        resdict[row] = ((env, result),)
        return self.add_results(resdict, allowdupe=allowdupe)


class ComposePage(ValidationPage):
    """A Page class that describes a single result page from a test
    compose or release candidate validation test event.
    """

    # I like this number, pylint
    # pylint: disable=too-many-arguments
    def __init__(self, site, release, testtype, milestone, compose, info=None, dist="Fedora"):
        super(ComposePage, self).__init__(
            site,
            release=release,
            milestone=milestone,
            compose=compose,
            testtype=testtype,
            info=info,
            dist=dist,
        )
        self.shortver = f"{self.milestone} {self.compose}"

        # String that will generate a clean copy of the page using the
        # test page generation template system.
        if self.shortdist:
            tmpl = f"{self.shortdist} validation results"
        else:
            tmpl = "Validation results"
        self.seedtext = (
            f"{{{{subst:{tmpl}|testtype={testtype}|release={self.release}|"
            f"milestone={self.milestone}|compose={self.compose}}}}}"
        )


class NightlyPage(ValidationPage):
    """A Page class that describes a single result page from a nightly
    validation test event.
    """

    # I like this number, pylint
    # pylint: disable=too-many-arguments
    def __init__(self, site, release, testtype, milestone, compose, info=None, dist="Fedora"):
        super(NightlyPage, self).__init__(
            site,
            release=release,
            milestone=milestone,
            compose=compose,
            testtype=testtype,
            info=info,
            dist=dist,
        )
        self.shortver = self.compose
        # overridden for nightlies to avoid expensive roundtrips
        if "." in compose:
            self.creation_date = compose.split(".")[0]
        else:
            self.creation_date = compose

        # String that will generate a clean copy of the page using the
        # test page generation template system.
        if self.shortdist:
            tmpl = f"{self.shortdist} validation results"
        else:
            tmpl = "Validation results"
        self.seedtext = (
            f"{{{{subst:{tmpl}|testtype={testtype}|release={self.release}|"
            f"milestone={self.milestone}|date={self.compose}}}}}"
        )


class SummaryPage(Page):
    """A Page class that describes the result summary page for a given
    event. event is the parent Event() for the page; summary pages are
    always considered to be a part of an Event.
    """

    def __init__(self, site, event, info=None, dist="Fedora"):
        self.dist = dist
        wikiname = f"Test Results:{dist} {event.version} Summary"
        self.shortdist = self.dist[7:]
        super(SummaryPage, self).__init__(site, wikiname, info)
        self.summary = f"Relval bot-created validation results summary for {dist} {event.version}"
        self.seedtext = (
            f"{dist} {event.version} [[QA:Release validation test plan|release "
            "validation]] summary. This page shows the results from all the "
            "individual result pages for this compose together. You can file "
            "results directly from this page and they will be saved into the "
            "correct individual result page. To see test instructions, visit "
            "any of the individual pages (the section titles are links). You "
            "can find download links below.\n\n"
        )
        self.seedtext += "__TOC__\n\n"
        self.seedtext += "== Downloads ==\n{{" + dist + " " + event.version + " Download}}"
        for testpage in event.valid_pages:
            self.seedtext += "\n\n== [[" + testpage.name + "|"
            self.seedtext += testpage.testtype + "]] ==\n{{"
            self.seedtext += testpage.name + "}}"

    def update_current(self):
        """Make the Current convenience redirect page on the wiki for the
        event point to this page.
        """
        if self.shortdist:
            curr = self.site.pages[f"Test Results:Current {self.shortdist} Summary"]
        else:
            curr = self.site.pages["Test Results:Current Summary"]
        curr.save(f"#REDIRECT [[{self.name}]]", "relval: update to current event", createonly=None)


class DownloadPage(Page):
    """The page containing image download links for a ValidationEvent.
    As with SummaryPage, is always associated with a specific event.
    """

    def __init__(self, site, event, info=None, dist="Fedora"):
        wikiname = f"Template:{dist} {event.version} Download"
        self.summary = f"Relval bot-created download page for {dist} {event.version}"
        super(DownloadPage, self).__init__(site, wikiname, info)
        self.event = event

    @property
    def seedtext(self):
        """A nicely formatted download table for the images for this
        compose. Here be dragons (and wiki table syntax). What you get
        from this is a table with one row for each unique 'image
        identifier' - the subvariant plus the image type - and columns
        for all arches in the entire image set; if there's an image
        for the given image type and arch then there'll be a download
        link in the appropriate column.
        """
        # pylint: disable=too-many-branches

        # sorting score values (see below)
        archscores = ((("x86_64", "i386"), 2000),)
        loadscores = (
            (("everything",), 300),
            (("workstation",), 220),
            (("server",), 210),
            (("cloud", "desktop", "cloud_base", "docker_base", "atomic"), 200),
            (("kde",), 190),
            (("minimal",), 90),
            (("xfce",), 80),
            (("soas",), 73),
            (("mate",), 72),
            (("cinnamon",), 71),
            (("lxde",), 70),
            (("source",), -10),
        )
        # Start by iterating over all images and grouping them by load
        # (that's imagedict) and keeping a record of each arch we
        # encounter (that's arches).
        arches = set()
        imagedict = OrderedDict()
        for img in self.event.ff_release_images.all_images:
            if img["arch"]:
                arches.add(img["arch"])
            # simple human-readable identifier for the image
            desc = " ".join((img["subvariant"], img["type"]))
            # assign a 'score' to the image; this will be used for
            # ordering the download table's rows.
            img["score"] = 0
            for (values, score) in archscores:
                if img["arch"] in values:
                    img["score"] = score
            for (values, score) in loadscores:
                if img["subvariant"].lower() in values:
                    img["score"] += score
            # The dict values are lists of images. We could use a
            # DefaultDict here, but faking it is easy too.
            if desc in imagedict:
                imagedict[desc].append(img)
            else:
                imagedict[desc] = [img]
        # Now we have our data, sort the dict using the weight we
        # calculated earlier. We use the max score of all arches in
        # each group of images.
        imagedict = OrderedDict(
            sorted(imagedict.items(), key=lambda x: max(img["score"] for img in x[1]), reverse=True)
        )
        # ...and sort the arches (just so they don't move around in
        # each new page and confuse people).
        arches = sorted(arches)

        # Now generate the table.
        table = '{| class="wikitable sortable mw-collapsible" width=100%\n|-\n'
        # Start of the header row...
        table += "! Image"
        for arch in arches:
            # Add a column for each arch
            table += f" !! {arch}"
        table += "\n"
        for (subvariant, imgs) in imagedict.items():
            # Add a row for each subvariant
            table += "|-\n"
            table += f"| {subvariant}\n"
            for arch in arches:
                # Add a cell for each arch (whether we have an image
                # or not)
                table += "| "
                for img in imgs:
                    if img["arch"] == arch:
                        # Add a link to the image if we have one
                        table += f"[{img['url']} Download]"
                table += "\n"
        # Close out the table when we're done
        table += "|-\n|}"
        return table

    def update_current(self):
        """Kind of a hack - relval needs this to exist as things
        stand. I'll probably refactor it later.
        """


class AMIPage(Page):
    """A page containing EC2 AMI links for a given event. Is included
    in the Cloud validation page to make it easy for people to find
    the correct AMIs.
    """

    def __init__(self, site, event, info=None, dist="Fedora"):
        wikiname = f"Template:{dist} {event.version} AMI"
        self.summary = f"Relval bot-created AMI page for {dist} {event.version}"
        super(AMIPage, self).__init__(site, wikiname, info)
        self.event = event

    @property
    def seedtext(self):
        """A table of all the AMIs for the compose for this event. We
        have to query this information out of datagrepper, that's the
        only place where it's available.
        """
        # pylint: disable=too-many-locals
        text = ""
        # first, let's get the information out of datagrepper. We'll
        # ask for messages up to 2 days after the event date.
        date = fedfind.helpers.parse_cid(self.event.ff_release.cid, dic=True)["date"]
        start = datetime.datetime.strptime(date, "%Y%m%d").replace(tzinfo=pytz.utc)
        end = start + datetime.timedelta(days=2)
        url = "https://apps.fedoraproject.org/datagrepper/raw"
        url += "?topic=org.fedoraproject.prod.fedimg.image.publish"
        # convert to epoch (this is what datagrepper wants)
        url += f"&start={start.timestamp()}&end={end.timestamp()}"
        json = fedfind.helpers.download_json(url)
        msgs = json["raw_messages"]
        # handle pagination
        for page in range(2, json["pages"] + 1):
            newurl = f"{url}&page={page}"
            newjson = fedfind.helpers.download_json(newurl)
            msgs.extend(newjson["raw_messages"])

        # now let's find the messages for our event compose
        ours = [msg["msg"] for msg in msgs if msg["msg"]["compose"] == self.event.ff_release.cid]

        def _table_line(msg):
            """Convenience function for generating a table line."""
            destination = msg["destination"]
            ami = msg["extra"]["id"]
            url = "https://redirect.fedoraproject.org/console.aws.amazon.com/ec2/v2/home?"
            url += f"region={destination}#LaunchInstanceWizard:ami={ami}"
            return f"| {destination}\n| {ami}\n| [{url} Launch in EC2]\n|-\n"

        def _table(arch, virttype, voltype):
            """Convenience function for adding a table."""
            ret = f"== {arch} {virttype} {voltype} AMIs ==\n\n"
            ret += '{| class="wikitable sortable mw-collapsible'
            if arch != "x86_64" or virttype != "hvm" or voltype != "standard":
                # we expand the x86_64 hvm standard table by default
                ret += " mw-collapsed"
            ret += '" width=100%\n|-\n'
            ret += "! Region !! AMI ID !! Direct launch link\n|-\n"
            # find the right messages for this arch and types
            relevants = [
                msg
                for msg in ours
                if msg["architecture"] == arch
                and msg["extra"]["virt_type"] == virttype
                and msg["extra"]["vol_type"] == voltype
            ]
            # sort the messages by region so the table is easier to scan
            relevants.sort(key=lambda x: x["destination"])
            for msg in relevants:
                ret += _table_line(msg)
            ret += "|}\n\n"
            return ret

        # now let's create and populate the tables
        for arch in ("x86_64", "arm64"):
            for virttype in ("hvm",):
                for voltype in ("standard", "gp2"):
                    text += _table(arch, virttype, voltype)

        return text


class TestDayPage(Page):
    """A Test Day results page. Usually contains table(s) with test
    cases as the column headers and users as the rows - each row is
    one user's results for all of the test cases in the table. Note
    this class is somewhat incomplete and really can only be used
    for its own methods, do *not* try writing one of these to the
    wiki.
    """

    def __init__(self, site, date, subject, info=None):
        # Handle names with no subject, e.g. Test_Day:2012-03-14
        wikiname = f"Test Day:{date}"
        if subject:
            wikiname = f"{wikiname} {subject}"
        super(TestDayPage, self).__init__(site, wikiname, info)
        self.date = date
        self.subject = subject
        self.results_separators = ("Test Results", "Results")

    def write(self, createonly=True):
        print("Creating Test Day pages is not yet supported.")

    @cached_property
    def bugs(self):
        """Returns a list of bug IDs referenced in the results section
        (as strings). Will find bugs in {{result}} and {{bz}}
        templates."""
        bugs = wikitcms.helpers.find_bugs(self.results_wikitext)
        for res in wikitcms.result.find_results_by_row(self.results_wikitext):
            bugs.update(res.bugs)
        return sorted(bugs)

    def fix_app_results(self):
        """The test day app does its own bug references outside the
        result template, instead of including them as the final
        parameters to the template like it should. This fixes that, in
        a fairly rough and ready way.
        """
        badres = re.compile(
            r"({{result.*?)}} {0,2}"
            r"(<ref>{{bz\|\d{6,7}}}</ref>) ?"
            r"(<ref>{{bz\|\d{6,7}}}</ref>)? ?"
            r"(<ref>{{bz\|\d{6,7}}}</ref>)? ?"
            r"(<ref>{{bz\|\d{6,7}}}</ref>)? ?"
            r"(<ref>{{bz\|\d{6,7}}}</ref>)? ?"
            r"(<ref>{{bz\|\d{6,7}}}</ref>)?"
        )
        text = oldtext = self.text()
        oldtext = text
        matches = badres.finditer(text)
        for match in matches:
            bugs = list()
            groups = match.groups()
            for group in groups[1:]:
                if group:
                    bugs.append(group[10:-8])
            text = text.replace(match.group(0), match.group(1) + "||" + "|".join(bugs) + "}}")
        return self.save(
            text,
            summary="Fix testday app-generated results to "
            "use {{result}} template for bug references",
            oldtext=oldtext,
        )

    def long_refs(self):
        """People tend to include giant essays as <ref> notes on test
        day results, which really makes table rendering ugly when
        they're dumped in the last column of the table. This finds all
        <ref> notes over 150 characters long, moves them to the "long"
        group, and adds a section at the end of the page with all the
        "long" notes in it. The 'end of page discovery' is a bit
        hacky, it just finds the last empty line in the page except
        for trailing lines and sticks the section there, but that's
        usually what we want - basically we want to make sure it
        appears just above the category memberships at the bottom of
        the page. It does go wrong *sometimes*, so good idea to check
        the page after it's edited.
        """
        text = oldtext = self.text()
        if '<ref group="long">' in text:
            # Don't run if we've already been run on this page
            return dict(nochange="")
        refpatt = re.compile("<ref>(.+?</ref>)", re.S)
        matches = refpatt.finditer(text)
        found = False
        for match in matches:
            if len(match.group(0)) > 150:
                found = True
                text = text.replace(match.group(0), '<ref group="long">' + match.group(1))
        if found:
            text = wikitcms.helpers.rreplace(
                text.strip(), "\n\n", '\n\n== Long comments ==\n<references group="long" />\n\n', 1
            )
            return self.save(
                text,
                summary="Move long comments to a separate " "section at end of page",
                oldtext=oldtext,
            )

        # If we didn't find any long refs, don't do anything
        return dict(nochange="")


# vim: set textwidth=100 ts=8 et sw=4:
